from pathlib import Path
from typing import Union, Text, Optional, Tuple

import matplotlib.pyplot as plt
from pyannote.core import Annotation, Segment, SlidingWindowFeature, notebook
from pyannote.database.util import load_rttm
from pyannote.metrics.diarization import DiarizationErrorRate
from rx.core import Observer
from typing_extensions import Literal

from . import utils


class WindowClosedException(Exception):
    pass


def _extract_prediction(value: Union[Tuple, Annotation]) -> Annotation:
    if isinstance(value, tuple):
        return value[0]
    if isinstance(value, Annotation):
        return value
    msg = f"Expected tuple or Annotation, but got {type(value)}"
    raise ValueError(msg)


class RTTMWriter(Observer):
    def __init__(self, uri: Text, path: Union[Path, Text], patch_collar: float = 0.05):
        super().__init__()
        self.uri = uri
        self.patch_collar = patch_collar
        self.path = Path(path).expanduser()
        if self.path.exists():
            self.path.unlink()

    def patch(self):
        """Stitch same-speaker turns that are close to each other"""
        if not self.path.exists():
            return
        annotations = list(load_rttm(self.path).values())
        if annotations:
            annotation = annotations[0]
            annotation.uri = self.uri
            with open(self.path, "w") as file:
                annotation.support(self.patch_collar).write_rttm(file)

    def on_next(self, value: Union[Tuple, Annotation]):
        prediction = _extract_prediction(value)
        # Write prediction in RTTM format
        prediction.uri = self.uri
        with open(self.path, "a") as file:
            prediction.write_rttm(file)

    def on_error(self, error: Exception):
        self.patch()

    def on_completed(self):
        self.patch()


class PredictionAccumulator(Observer):
    def __init__(self, uri: Optional[Text] = None, patch_collar: float = 0.05):
        super().__init__()
        self.uri = uri
        self.patch_collar = patch_collar
        self._prediction: Optional[Annotation] = None

    def patch(self):
        """Stitch same-speaker turns that are close to each other"""
        if self._prediction is not None:
            self._prediction = self._prediction.support(self.patch_collar)

    def get_prediction(self) -> Annotation:
        # Patch again in case this is called before on_completed
        self.patch()
        return self._prediction

    def on_next(self, value: Union[Tuple, Annotation]):
        prediction = _extract_prediction(value)
        prediction.uri = self.uri
        if self._prediction is None:
            self._prediction = prediction
        else:
            self._prediction.update(prediction)

    def on_error(self, error: Exception):
        self.patch()

    def on_completed(self):
        self.patch()


class StreamingPlot(Observer):
    def __init__(
        self,
        duration: float,
        latency: float,
        visualization: Literal["slide", "accumulate"] = "slide",
        reference: Optional[Union[Path, Text]] = None,
    ):
        super().__init__()
        assert visualization in ["slide", "accumulate"]
        self.visualization = visualization
        self.reference = reference
        if self.reference is not None:
            self.reference = list(load_rttm(reference).values())[0]
        self.window_duration = duration
        self.latency = latency
        self.figure, self.axs, self.num_axs = None, None, -1
        # This flag allows to catch the matplotlib window closed event and make the next call stop iterating
        self.window_closed = False

    def _on_window_closed(self, event):
        self.window_closed = True

    def _init_num_axs(self):
        if self.num_axs == -1:
            self.num_axs = 2
            if self.reference is not None:
                self.num_axs += 1

    def _init_figure(self):
        self._init_num_axs()
        self.figure, self.axs = plt.subplots(
            self.num_axs, 1, figsize=(10, 2 * self.num_axs)
        )
        if self.num_axs == 1:
            self.axs = [self.axs]
        self.figure.canvas.mpl_connect("close_event", self._on_window_closed)

    def _clear_axs(self):
        for i in range(self.num_axs):
            self.axs[i].clear()

    def get_plot_bounds(self, real_time: float) -> Segment:
        start_time = 0
        end_time = real_time - self.latency
        if self.visualization == "slide":
            start_time = max(0.0, end_time - self.window_duration)
        return Segment(start_time, end_time)

    def on_next(self, values: Tuple[Annotation, SlidingWindowFeature, float]):
        if self.window_closed:
            raise WindowClosedException

        prediction, waveform, real_time = values

        # Initialize figure if first call
        if self.figure is None:
            self._init_figure()
        # Clear previous plots
        self._clear_axs()
        # Set plot bounds
        notebook.crop = self.get_plot_bounds(real_time)

        # Align prediction and reference if possible
        if self.reference is not None:
            metric = DiarizationErrorRate()
            mapping = metric.optimal_mapping(self.reference, prediction)
            prediction.rename_labels(mapping=mapping, copy=False)

        # Plot prediction
        notebook.plot_annotation(prediction, self.axs[0])
        self.axs[0].set_title("Output")

        # Plot waveform
        notebook.plot_feature(waveform, self.axs[1])
        self.axs[1].set_title("Audio")

        # Plot reference if available
        if self.num_axs == 3:
            notebook.plot_annotation(self.reference, self.axs[2])
            self.axs[2].set_title("Reference")

        # Draw
        plt.tight_layout()
        self.figure.canvas.draw()
        self.figure.canvas.flush_events()
        plt.pause(0.05)



import argparse
from pathlib import Path
from threading import Thread
from typing import Text, Optional

import rx.operators as ops
from websocket import WebSocket

from diart import argdoc
from diart import sources as src
from diart import utils


def send_audio(ws: WebSocket, source: Text, step: float, sample_rate: int):
    # Create audio source
    source_components = source.split(":")
    if source_components[0] != "microphone":
        audio_source = src.FileAudioSource(source, sample_rate, block_duration=step)
    else:
        device = int(source_components[1]) if len(source_components) > 1 else None
        audio_source = src.MicrophoneAudioSource(step, device)

    # Encode audio, then send through websocket
    audio_source.stream.pipe(ops.map(utils.encode_audio)).subscribe_(ws.send)

    # Start reading audio
    audio_source.read()


def receive_audio(ws: WebSocket, output: Optional[Path]):
    while True:
        message = ws.recv()
        print(f"Received: {message}", end="")
        if output is not None:
            with open(output, "a") as file:
                file.write(message)


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "source",
        type=str,
        help="Path to an audio file | 'microphone' | 'microphone:<DEVICE_ID>'",
    )
    parser.add_argument("--host", required=True, type=str, help="Server host")
    parser.add_argument("--port", required=True, type=int, help="Server port")
    parser.add_argument(
        "--step", default=0.5, type=float, help=f"{argdoc.STEP}. Defaults to 0.5"
    )
    parser.add_argument(
        "-sr",
        "--sample-rate",
        default=16000,
        type=int,
        help=f"{argdoc.SAMPLE_RATE}. Defaults to 16000",
    )
    parser.add_argument(
        "-o",
        "--output-file",
        type=Path,
        help="Output RTTM file. Defaults to no writing",
    )
    args = parser.parse_args()

    # Run websocket client
    ws = WebSocket()
    ws.connect(f"ws://{args.host}:{args.port}")
    sender = Thread(
        target=send_audio, args=[ws, args.source, args.step, args.sample_rate]
    )
    receiver = Thread(target=receive_audio, args=[ws, args.output_file])
    sender.start()
    receiver.start()


if __name__ == "__main__":
    run()



import argparse
from pathlib import Path

import torch

from diart import argdoc
from diart import models as m
from diart import sources as src
from diart import utils
from diart.inference import StreamingInference
from diart.sinks import RTTMWriter


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "source",
        type=str,
        help="Path to an audio file | 'microphone' | 'microphone:<DEVICE_ID>'",
    )
    parser.add_argument(
        "--pipeline",
        default="SpeakerDiarization",
        type=str,
        help="Class of the pipeline to optimize. Defaults to 'SpeakerDiarization'",
    )
    parser.add_argument(
        "--segmentation",
        default="pyannote/segmentation",
        type=str,
        help=f"{argdoc.SEGMENTATION}. Defaults to pyannote/segmentation",
    )
    parser.add_argument(
        "--embedding",
        default="pyannote/embedding",
        type=str,
        help=f"{argdoc.EMBEDDING}. Defaults to pyannote/embedding",
    )
    parser.add_argument(
        "--duration",
        type=float,
        help=f"{argdoc.DURATION}. Defaults to training segmentation duration",
    )
    parser.add_argument(
        "--step", default=0.5, type=float, help=f"{argdoc.STEP}. Defaults to 0.5"
    )
    parser.add_argument(
        "--latency", default=0.5, type=float, help=f"{argdoc.LATENCY}. Defaults to 0.5"
    )
    parser.add_argument(
        "--tau-active", default=0.5, type=float, help=f"{argdoc.TAU}. Defaults to 0.5"
    )
    parser.add_argument(
        "--rho-update", default=0.3, type=float, help=f"{argdoc.RHO}. Defaults to 0.3"
    )
    parser.add_argument(
        "--delta-new", default=1, type=float, help=f"{argdoc.DELTA}. Defaults to 1"
    )
    parser.add_argument(
        "--gamma", default=3, type=float, help=f"{argdoc.GAMMA}. Defaults to 3"
    )
    parser.add_argument(
        "--beta", default=10, type=float, help=f"{argdoc.BETA}. Defaults to 10"
    )
    parser.add_argument(
        "--max-speakers",
        default=20,
        type=int,
        help=f"{argdoc.MAX_SPEAKERS}. Defaults to 20",
    )
    parser.add_argument(
        "--no-plot",
        dest="no_plot",
        action="store_true",
        help="Skip plotting for faster inference",
    )
    parser.add_argument(
        "--cpu",
        dest="cpu",
        action="store_true",
        help=f"{argdoc.CPU}. Defaults to GPU if available, CPU otherwise",
    )
    parser.add_argument(
        "--output",
        type=str,
        help=f"{argdoc.OUTPUT}. Defaults to home directory if SOURCE == 'microphone' or parent directory if SOURCE is a file",
    )
    parser.add_argument(
        "--hf-token",
        default="true",
        type=str,
        help=f"{argdoc.HF_TOKEN}. Defaults to 'true' (required by pyannote)",
    )
    args = parser.parse_args()

    # Resolve device
    args.device = torch.device("cpu") if args.cpu else None

    # Resolve models
    hf_token = utils.parse_hf_token_arg(args.hf_token)
    args.segmentation = m.SegmentationModel.from_pyannote(args.segmentation, hf_token)
    args.embedding = m.EmbeddingModel.from_pyannote(args.embedding, hf_token)

    # Resolve pipeline
    pipeline_class = utils.get_pipeline_class(args.pipeline)
    config = pipeline_class.get_config_class()(**vars(args))
    pipeline = pipeline_class(config)

    # Manage audio source
    source_components = args.source.split(":")
    if source_components[0] != "microphone":
        args.source = Path(args.source).expanduser()
        args.output = args.source.parent if args.output is None else Path(args.output)
        padding = config.get_file_padding(args.source)
        audio_source = src.FileAudioSource(
            args.source, config.sample_rate, padding, config.step
        )
        pipeline.set_timestamp_shift(-padding[0])
    else:
        args.output = (
            Path("~/").expanduser() if args.output is None else Path(args.output)
        )
        device = int(source_components[1]) if len(source_components) > 1 else None
        audio_source = src.MicrophoneAudioSource(config.step, device)

    # Run online inference
    inference = StreamingInference(
        pipeline,
        audio_source,
        batch_size=1,
        do_profile=True,
        do_plot=not args.no_plot,
        show_progress=True,
    )
    inference.attach_observers(
        RTTMWriter(audio_source.uri, args.output / f"{audio_source.uri}.rttm")
    )
    try:
        inference()
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    run()



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Tuple, Sequence, Text

from pyannote.core import SlidingWindowFeature
from pyannote.metrics.base import BaseMetric

from .. import utils
from ..audio import FilePath, AudioLoader


@dataclass
class HyperParameter:
    name: Text
    low: float
    high: float

    @staticmethod
    def from_name(name: Text) -> "HyperParameter":
        if name == "tau_active":
            return TauActive
        if name == "rho_update":
            return RhoUpdate
        if name == "delta_new":
            return DeltaNew
        raise ValueError(f"Hyper-parameter '{name}' not recognized")


TauActive = HyperParameter("tau_active", low=0, high=1)
RhoUpdate = HyperParameter("rho_update", low=0, high=1)
DeltaNew = HyperParameter("delta_new", low=0, high=2)


class PipelineConfig(ABC):
    @property
    @abstractmethod
    def duration(self) -> float:
        pass

    @property
    @abstractmethod
    def step(self) -> float:
        pass

    @property
    @abstractmethod
    def latency(self) -> float:
        pass

    @property
    @abstractmethod
    def sample_rate(self) -> int:
        pass

    def get_file_padding(self, filepath: FilePath) -> Tuple[float, float]:
        file_duration = AudioLoader(self.sample_rate, mono=True).get_duration(filepath)
        right = utils.get_padding_right(self.latency, self.step)
        left = utils.get_padding_left(file_duration + right, self.duration)
        return left, right


class Pipeline(ABC):
    @staticmethod
    @abstractmethod
    def get_config_class() -> type:
        pass

    @staticmethod
    @abstractmethod
    def suggest_metric() -> BaseMetric:
        pass

    @staticmethod
    @abstractmethod
    def hyper_parameters() -> Sequence[HyperParameter]:
        pass

    @property
    @abstractmethod
    def config(self) -> PipelineConfig:
        pass

    @abstractmethod
    def reset(self):
        pass

    @abstractmethod
    def set_timestamp_shift(self, shift: float):
        pass

    @abstractmethod
    def __call__(
        self, waveforms: Sequence[SlidingWindowFeature]
    ) -> Sequence[Tuple[Any, SlidingWindowFeature]]:
        pass



from collections import OrderedDict
from pathlib import Path
from typing import Sequence, Text, Optional, Union

from optuna import TrialPruned, Study, create_study
from optuna.samplers import TPESampler
from optuna.trial import Trial, FrozenTrial
from pyannote.metrics.base import BaseMetric
from tqdm import trange, tqdm
from typing_extensions import Literal

from . import blocks
from .audio import FilePath
from .inference import Benchmark


class Optimizer:
    def __init__(
        self,
        pipeline_class: type,
        speech_path: Union[Text, Path],
        reference_path: Union[Text, Path],
        study_or_path: Union[FilePath, Study],
        batch_size: int = 32,
        hparams: Optional[Sequence[blocks.base.HyperParameter]] = None,
        base_config: Optional[blocks.PipelineConfig] = None,
        do_kickstart_hparams: bool = True,
        metric: Optional[BaseMetric] = None,
        direction: Literal["minimize", "maximize"] = "minimize",
    ):
        self.pipeline_class = pipeline_class
        # FIXME can we run this benchmark in parallel?
        #  Currently it breaks the trial progress bar
        self.benchmark = Benchmark(
            speech_path,
            reference_path,
            show_progress=True,
            show_report=False,
            batch_size=batch_size,
        )

        self.metric = metric
        self.direction = direction
        self.base_config = base_config
        self.do_kickstart_hparams = do_kickstart_hparams
        if self.base_config is None:
            self.base_config = self.pipeline_class.get_config_class()()
            self.do_kickstart_hparams = False

        self.hparams = hparams
        if self.hparams is None:
            self.hparams = self.pipeline_class.hyper_parameters()

        # Make sure hyper-parameters exist in the configuration class given
        possible_hparams = vars(self.base_config)
        for param in self.hparams:
            msg = (
                f"Hyper-parameter {param.name} not found "
                f"in configuration {self.base_config.__class__.__name__}"
            )
            assert param.name in possible_hparams, msg

        self._progress: Optional[tqdm] = None

        if isinstance(study_or_path, Study):
            self.study = study_or_path
        elif isinstance(study_or_path, str) or isinstance(study_or_path, Path):
            study_or_path = Path(study_or_path)
            self.study = create_study(
                storage="sqlite:///" + str(study_or_path / f"{study_or_path.stem}.db"),
                sampler=TPESampler(),
                study_name=study_or_path.stem,
                direction=self.direction,
                load_if_exists=True,
            )
        else:
            msg = f"Expected Study object or path-like, but got {type(study_or_path).__name__}"
            raise ValueError(msg)

    @property
    def best_performance(self):
        return self.study.best_value

    @property
    def best_hparams(self):
        return self.study.best_params

    def _callback(self, study: Study, trial: FrozenTrial):
        if self._progress is None:
            return
        self._progress.update(1)
        self._progress.set_description(f"Trial {trial.number + 1}")
        values = {"best_perf": study.best_value}
        for name, value in study.best_params.items():
            values[f"best_{name}"] = value
        self._progress.set_postfix(OrderedDict(values))

    def objective(self, trial: Trial) -> float:
        # Set suggested values for optimized hyper-parameters
        trial_config = vars(self.base_config)
        for hparam in self.hparams:
            trial_config[hparam.name] = trial.suggest_uniform(
                hparam.name, hparam.low, hparam.high
            )

        # Prune trial if required
        if trial.should_prune():
            raise TrialPruned()

        # Instantiate the new configuration for the trial
        config = self.base_config.__class__(**trial_config)

        # Determine the evaluation metric
        metric = self.metric
        if metric is None:
            metric = self.pipeline_class.suggest_metric()

        # Run pipeline over the dataset
        report = self.benchmark(self.pipeline_class, config, metric)

        # Extract target metric from report
        return report.loc["TOTAL", metric.name]["%"]

    def __call__(self, num_iter: int, show_progress: bool = True):
        self._progress = None
        if show_progress:
            self._progress = trange(num_iter)
            last_trial = -1
            if self.study.trials:
                last_trial = self.study.trials[-1].number
            self._progress.set_description(f"Trial {last_trial + 1}")
        # Start with base config hyper-parameters if config was given
        if self.do_kickstart_hparams:
            self.study.enqueue_trial(
                {
                    param.name: getattr(self.base_config, param.name)
                    for param in self.hparams
                },
                skip_if_exists=True,
            )
        self.study.optimize(self.objective, num_iter, callbacks=[self._callback])



import base64
import time
from typing import Optional, Text, Union

import matplotlib.pyplot as plt
import numpy as np
from pyannote.core import Annotation, Segment, SlidingWindowFeature, notebook

from . import blocks
from .progress import ProgressBar


class Chronometer:
    def __init__(self, unit: Text, progress_bar: Optional[ProgressBar] = None):
        self.unit = unit
        self.progress_bar = progress_bar
        self.current_start_time = None
        self.history = []

    @property
    def is_running(self):
        return self.current_start_time is not None

    def start(self):
        self.current_start_time = time.monotonic()

    def stop(self, do_count: bool = True):
        msg = "No start time available, Did you call stop() before start()?"
        assert self.current_start_time is not None, msg
        end_time = time.monotonic() - self.current_start_time
        self.current_start_time = None
        if do_count:
            self.history.append(end_time)

    def report(self):
        print_fn = print
        if self.progress_bar is not None:
            print_fn = self.progress_bar.write
        print_fn(
            f"Took {np.mean(self.history).item():.3f} "
            f"(+/-{np.std(self.history).item():.3f}) seconds/{self.unit} "
            f"-- ran {len(self.history)} times"
        )


def parse_hf_token_arg(hf_token: Union[bool, Text]) -> Union[bool, Text]:
    if isinstance(hf_token, bool):
        return hf_token
    if hf_token.lower() == "true":
        return True
    if hf_token.lower() == "false":
        return False
    return hf_token


def encode_audio(waveform: np.ndarray) -> Text:
    data = waveform.astype(np.float32).tobytes()
    return base64.b64encode(data).decode("utf-8")


def decode_audio(data: Text) -> np.ndarray:
    # Decode chunk encoded in base64
    byte_samples = base64.decodebytes(data.encode("utf-8"))
    # Recover array from bytes
    samples = np.frombuffer(byte_samples, dtype=np.float32)
    return samples.reshape(1, -1)


def get_padding_left(stream_duration: float, chunk_duration: float) -> float:
    if stream_duration < chunk_duration:
        return chunk_duration - stream_duration
    return 0


def repeat_label(label: Text):
    while True:
        yield label


def get_pipeline_class(class_name: Text) -> type:
    pipeline_class = getattr(blocks, class_name, None)
    msg = f"Pipeline '{class_name}' doesn't exist"
    assert pipeline_class is not None, msg
    return pipeline_class


def get_padding_right(latency: float, step: float) -> float:
    return latency - step


def visualize_feature(duration: Optional[float] = None):
    def apply(feature: SlidingWindowFeature):
        if duration is None:
            notebook.crop = feature.extent
        else:
            notebook.crop = Segment(feature.extent.end - duration, feature.extent.end)
        plt.rcParams["figure.figsize"] = (8, 2)
        notebook.plot_feature(feature)
        plt.tight_layout()
        plt.show()

    return apply


def visualize_annotation(duration: Optional[float] = None):
    def apply(annotation: Annotation):
        extent = annotation.get_timeline().extent()
        if duration is None:
            notebook.crop = extent
        else:
            notebook.crop = Segment(extent.end - duration, extent.end)
        plt.rcParams["figure.figsize"] = (8, 2)
        notebook.plot_annotation(annotation)
        plt.tight_layout()
        plt.show()

    return apply



from typing import Optional, List, Iterable, Tuple

import numpy as np
import torch
from pyannote.core import SlidingWindowFeature

from ..mapping import SpeakerMap, SpeakerMapBuilder


class OnlineSpeakerClustering:
    """Implements constrained incremental online clustering of speakers and manages cluster centers.

    Parameters
    ----------
    tau_active:float
        Threshold for detecting active speakers. This threshold is applied on the maximum value of per-speaker output
        activation of the local segmentation model.
    rho_update: float
        Threshold for considering the extracted embedding when updating the centroid of the local speaker.
        The centroid to which a local speaker is mapped is only updated if the ratio of speech/chunk duration
        of a given local speaker is greater than this threshold.
    delta_new: float
        Threshold on the distance between a speaker embedding and a centroid. If the distance between a local speaker and all
        centroids is larger than delta_new, then a new centroid is created for the current speaker.
    metric: str. Defaults to "cosine".
        The distance metric to use.
    max_speakers: int
        Maximum number of global speakers to track through a conversation. Defaults to 20.
    """

    def __init__(
        self,
        tau_active: float,
        rho_update: float,
        delta_new: float,
        metric: Optional[str] = "cosine",
        max_speakers: int = 20,
    ):
        self.tau_active = tau_active
        self.rho_update = rho_update
        self.delta_new = delta_new
        self.metric = metric
        self.max_speakers = max_speakers
        self.centers: Optional[np.ndarray] = None
        self.active_centers = set()
        self.blocked_centers = set()

    @property
    def num_free_centers(self) -> int:
        return self.max_speakers - self.num_known_speakers - self.num_blocked_speakers

    @property
    def num_known_speakers(self) -> int:
        return len(self.active_centers)

    @property
    def num_blocked_speakers(self) -> int:
        return len(self.blocked_centers)

    @property
    def inactive_centers(self) -> List[int]:
        return [
            c
            for c in range(self.max_speakers)
            if c not in self.active_centers or c in self.blocked_centers
        ]

    def get_next_center_position(self) -> Optional[int]:
        for center in range(self.max_speakers):
            if center not in self.active_centers and center not in self.blocked_centers:
                return center

    def init_centers(self, dimension: int):
        """Initializes the speaker centroid matrix

        Parameters
        ----------
        dimension: int
            Dimension of embeddings used for representing a speaker.
        """
        self.centers = np.zeros((self.max_speakers, dimension))
        self.active_centers = set()
        self.blocked_centers = set()

    def update(self, assignments: Iterable[Tuple[int, int]], embeddings: np.ndarray):
        """Updates the speaker centroids given a list of assignments and local speaker embeddings

        Parameters
        ----------
        assignments: Iterable[Tuple[int, int]])
            An iterable of tuples with two elements having the first element as the source speaker
            and the second element as the target speaker.
        embeddings: np.ndarray, shape (local_speakers, embedding_dim)
            Matrix containing embeddings for all local speakers.
        """
        if self.centers is not None:
            for l_spk, g_spk in assignments:
                assert g_spk in self.active_centers, "Cannot update unknown centers"
                self.centers[g_spk] += embeddings[l_spk]

    def add_center(self, embedding: np.ndarray) -> int:
        """Add a new speaker centroid initialized to a given embedding

        Parameters
        ----------
        embedding: np.ndarray
            Embedding vector of some local speaker

        Returns
        -------
            center_index: int
                Index of the created center
        """
        center = self.get_next_center_position()
        self.centers[center] = embedding
        self.active_centers.add(center)
        return center

    def identify(
        self, segmentation: SlidingWindowFeature, embeddings: torch.Tensor
    ) -> SpeakerMap:
        """Identify the centroids to which the input speaker embeddings belong.

        Parameters
        ----------
        segmentation: np.ndarray, shape (frames, local_speakers)
            Matrix of segmentation outputs
        embeddings: np.ndarray, shape (local_speakers, embedding_dim)
            Matrix of embeddings

        Returns
        -------
            speaker_map: SpeakerMap
                A mapping from local speakers to global speakers.
        """
        embeddings = embeddings.detach().cpu().numpy()
        active_speakers = np.where(
            np.max(segmentation.data, axis=0) >= self.tau_active
        )[0]
        long_speakers = np.where(np.mean(segmentation.data, axis=0) >= self.rho_update)[
            0
        ]
        num_local_speakers = segmentation.data.shape[1]

        if self.centers is None:
            self.init_centers(embeddings.shape[1])
            assignments = [
                (spk, self.add_center(embeddings[spk])) for spk in active_speakers
            ]
            return SpeakerMapBuilder.hard_map(
                shape=(num_local_speakers, self.max_speakers),
                assignments=assignments,
                maximize=False,
            )

        # Obtain a mapping based on distances between embeddings and centers
        dist_map = SpeakerMapBuilder.dist(embeddings, self.centers, self.metric)
        # Remove any assignments containing invalid speakers
        inactive_speakers = np.array(
            [spk for spk in range(num_local_speakers) if spk not in active_speakers]
        )
        dist_map = dist_map.unmap_speakers(inactive_speakers, self.inactive_centers)
        # Keep assignments under the distance threshold
        valid_map = dist_map.unmap_threshold(self.delta_new)

        # Some speakers might be unidentified
        missed_speakers = [
            s for s in active_speakers if not valid_map.is_source_speaker_mapped(s)
        ]

        # Add assignments to new centers if possible
        new_center_speakers = []
        for spk in missed_speakers:
            has_space = len(new_center_speakers) < self.num_free_centers
            if has_space and spk in long_speakers:
                # Flag as a new center
                new_center_speakers.append(spk)
            else:
                # Cannot create a new center
                # Get global speakers in order of preference
                preferences = np.argsort(dist_map.mapping_matrix[spk, :])
                preferences = [
                    g_spk for g_spk in preferences if g_spk in self.active_centers
                ]
                # Get the free global speakers among the preferences
                _, g_assigned = valid_map.valid_assignments()
                free = [g_spk for g_spk in preferences if g_spk not in g_assigned]
                if free:
                    # The best global speaker is the closest free one
                    valid_map = valid_map.set_source_speaker(spk, free[0])

        # Update known centers
        to_update = [
            (ls, gs)
            for ls, gs in zip(*valid_map.valid_assignments())
            if ls not in missed_speakers and ls in long_speakers
        ]
        self.update(to_update, embeddings)

        # Add new centers
        for spk in new_center_speakers:
            valid_map = valid_map.set_source_speaker(
                spk, self.add_center(embeddings[spk])
            )

        return valid_map

    def __call__(
        self, segmentation: SlidingWindowFeature, embeddings: torch.Tensor
    ) -> SlidingWindowFeature:
        return SlidingWindowFeature(
            self.identify(segmentation, embeddings).apply(segmentation.data),
            segmentation.sliding_window,
        )



from typing import Text, Optional

import numpy as np
import torch
from pyannote.core import Annotation, Segment, SlidingWindowFeature
import torchaudio.transforms as T

from ..features import TemporalFeatures, TemporalFeatureFormatter


class Binarize:
    """
    Transform a speaker segmentation from the discrete-time domain
    into a continuous-time speaker segmentation.

    Parameters
    ----------
    threshold: float
        Probability threshold to determine if a speaker is active at a given frame.
    uri: Optional[Text]
        Uri of the audio stream. Defaults to no uri.
    """

    def __init__(self, threshold: float, uri: Optional[Text] = None):
        self.uri = uri
        self.threshold = threshold

    def __call__(self, segmentation: SlidingWindowFeature) -> Annotation:
        """
        Return the continuous-time segmentation
        corresponding to the discrete-time input segmentation.

        Parameters
        ----------
        segmentation: SlidingWindowFeature
            Discrete-time speaker segmentation.

        Returns
        -------
        annotation: Annotation
            Continuous-time speaker segmentation.
        """
        num_frames, num_speakers = segmentation.data.shape
        timestamps = segmentation.sliding_window
        is_active = segmentation.data > self.threshold
        # Artificially add last inactive frame to close any remaining speaker turns
        is_active = np.append(is_active, [[False] * num_speakers], axis=0)
        start_times = np.zeros(num_speakers) + timestamps[0].middle
        annotation = Annotation(uri=self.uri, modality="speech")
        for t in range(num_frames):
            # Any (False, True) starts a speaker turn at "True" index
            onsets = np.logical_and(np.logical_not(is_active[t]), is_active[t + 1])
            start_times[onsets] = timestamps[t + 1].middle
            # Any (True, False) ends a speaker turn at "False" index
            offsets = np.logical_and(is_active[t], np.logical_not(is_active[t + 1]))
            for spk in np.where(offsets)[0]:
                region = Segment(start_times[spk], timestamps[t + 1].middle)
                annotation[region, spk] = f"speaker{spk}"
        return annotation


class Resample:
    """Dynamically resample audio chunks.

    Parameters
    ----------
    sample_rate: int
        Original sample rate of the input audio
    resample_rate: int
        Sample rate of the output
    """

    def __init__(
        self,
        sample_rate: int,
        resample_rate: int,
        device: Optional[torch.device] = None,
    ):
        self.device = device
        if self.device is None:
            self.device = torch.device("cpu")
        self.resample = T.Resample(sample_rate, resample_rate).to(self.device)
        self.formatter = TemporalFeatureFormatter()

    def __call__(self, waveform: TemporalFeatures) -> TemporalFeatures:
        wav = self.formatter.cast(waveform).to(self.device)  # shape (batch, samples, 1)
        with torch.no_grad():
            resampled_wav = self.resample(wav.transpose(-1, -2)).transpose(-1, -2)
        return self.formatter.restore_type(resampled_wav)


class AdjustVolume:
    """Change the volume of an audio chunk.

    Notice that the output volume might be different to avoid saturation.

    Parameters
    ----------
    volume_in_db: float
        Target volume in dB.
    """

    def __init__(self, volume_in_db: float):
        self.target_db = volume_in_db
        self.formatter = TemporalFeatureFormatter()

    @staticmethod
    def get_volumes(waveforms: torch.Tensor) -> torch.Tensor:
        """Compute the volumes of a set of audio chunks.

        Parameters
        ----------
        waveforms: torch.Tensor
            Audio chunks. Shape (batch, samples, channels).

        Returns
        -------
        volumes: torch.Tensor
            Audio chunk volumes per channel. Shape (batch, 1, channels)
        """
        return 10 * torch.log10(
            torch.mean(torch.abs(waveforms) ** 2, dim=1, keepdim=True)
        )

    def __call__(self, waveform: TemporalFeatures) -> TemporalFeatures:
        wav = self.formatter.cast(waveform)  # shape (batch, samples, channels)
        with torch.no_grad():
            # Compute current volume per chunk, shape (batch, 1, channels)
            current_volumes = self.get_volumes(wav)
            # Determine gain to reach the target volume
            gains = 10 ** ((self.target_db - current_volumes) / 20)
            # Apply gain
            wav = gains * wav
            # If maximum value is greater than one, normalize chunk
            maximums = torch.clamp(torch.amax(torch.abs(wav), dim=1, keepdim=True), 1)
            wav = wav / maximums
        return self.formatter.restore_type(wav)



import logging
from multiprocessing import Pool, freeze_support, RLock, current_process
from pathlib import Path
from traceback import print_exc
from typing import Union, Text, Optional, Callable, Tuple, List

import numpy as np
import pandas as pd
import rx
import rx.operators as ops
import torch
from pyannote.core import Annotation, SlidingWindowFeature
from pyannote.database.util import load_rttm
from pyannote.metrics.base import BaseMetric
from rx.core import Observer
from tqdm import tqdm

from . import blocks
from . import operators as dops
from . import sources as src
from . import utils
from .progress import ProgressBar, RichProgressBar, TQDMProgressBar
from .sinks import PredictionAccumulator, StreamingPlot, WindowClosedException


class StreamingInference:
    """Performs inference in real time given a pipeline and an audio source.
    Streams an audio source to an online speaker diarization pipeline.
    It allows users to attach a chain of operations in the form of hooks.

    Parameters
    ----------
    pipeline: StreamingPipeline
        Configured speaker diarization pipeline.
    source: AudioSource
        Audio source to be read and streamed.
    batch_size: int
        Number of inputs to send to the pipeline at once.
        Defaults to 1.
    do_profile: bool
        If True, compute and report the processing time of the pipeline.
        Defaults to True.
    do_plot: bool
        If True, draw predictions in a moving plot.
        Defaults to False.
    show_progress: bool
        If True, show a progress bar.
        Defaults to True.
    progress_bar: Optional[diart.progress.ProgressBar]
        Progress bar.
        If description is not provided, set to 'Streaming <source uri>'.
        Defaults to RichProgressBar().
    """

    def __init__(
        self,
        pipeline: blocks.Pipeline,
        source: src.AudioSource,
        batch_size: int = 1,
        do_profile: bool = True,
        do_plot: bool = False,
        show_progress: bool = True,
        progress_bar: Optional[ProgressBar] = None,
    ):
        self.pipeline = pipeline
        self.source = source
        self.batch_size = batch_size
        self.do_profile = do_profile
        self.do_plot = do_plot
        self.show_progress = show_progress
        self.accumulator = PredictionAccumulator(self.source.uri)
        self.unit = "chunk" if self.batch_size == 1 else "batch"
        self._observers = []

        chunk_duration = self.pipeline.config.duration
        step_duration = self.pipeline.config.step
        sample_rate = self.pipeline.config.sample_rate

        # Estimate the total number of chunks that the source will emit
        self.num_chunks = None
        if self.source.duration is not None:
            numerator = self.source.duration - chunk_duration + step_duration
            self.num_chunks = int(np.ceil(numerator / step_duration))

        # Show progress if required
        self._pbar = progress_bar
        if self.show_progress:
            if self._pbar is None:
                self._pbar = RichProgressBar()
            self._pbar.create(
                total=self.num_chunks,
                description=f"Streaming {self.source.uri}",
                unit=self.unit,
            )

        # Initialize chronometer for profiling
        self._chrono = utils.Chronometer(self.unit, self._pbar)

        self.stream = self.source.stream

        # Rearrange stream to form sliding windows
        self.stream = self.stream.pipe(
            dops.rearrange_audio_stream(
                chunk_duration, step_duration, source.sample_rate
            ),
        )

        # Dynamic resampling if the audio source isn't compatible
        if sample_rate != self.source.sample_rate:
            msg = (
                f"Audio source has sample rate {self.source.sample_rate}, "
                f"but pipeline's is {sample_rate}. Will resample."
            )
            logging.warning(msg)
            self.stream = self.stream.pipe(
                ops.map(
                    blocks.Resample(
                        self.source.sample_rate,
                        sample_rate,
                        self.pipeline.config.device,
                    )
                )
            )

        # Form batches
        self.stream = self.stream.pipe(
            ops.buffer_with_count(count=self.batch_size),
        )

        if self.do_profile:
            self.stream = self.stream.pipe(
                ops.do_action(on_next=lambda _: self._chrono.start()),
                ops.map(self.pipeline),
                ops.do_action(on_next=lambda _: self._chrono.stop()),
            )
        else:
            self.stream = self.stream.pipe(ops.map(self.pipeline))

        self.stream = self.stream.pipe(
            ops.flat_map(lambda results: rx.from_iterable(results)),
            ops.do(self.accumulator),
        )

        if show_progress:
            self.stream = self.stream.pipe(
                ops.do_action(on_next=lambda _: self._pbar.update())
            )

    def _close_pbar(self):
        if self._pbar is not None:
            self._pbar.close()

    def _close_chronometer(self):
        if self.do_profile:
            if self._chrono.is_running:
                self._chrono.stop(do_count=False)
            self._chrono.report()

    def attach_hooks(
        self, *hooks: Callable[[Tuple[Annotation, SlidingWindowFeature]], None]
    ):
        """Attach hooks to the pipeline.

        Parameters
        ----------
        *hooks: (Tuple[Annotation, SlidingWindowFeature]) -> None
            Hook functions to consume emitted annotations and audio.
        """
        self.stream = self.stream.pipe(*[ops.do_action(hook) for hook in hooks])

    def attach_observers(self, *observers: Observer):
        """Attach rx observers to the pipeline.

        Parameters
        ----------
        *observers: Observer
            Observers to consume emitted annotations and audio.
        """
        self.stream = self.stream.pipe(*[ops.do(sink) for sink in observers])
        self._observers.extend(observers)

    def _handle_error(self, error: BaseException):
        # Compensate for Rx not always calling on_error
        for sink in self._observers:
            sink.on_error(error)
        # Always close the source in case of bad termination
        self.source.close()
        # Special treatment for a user interruption (counted as normal termination)
        window_closed = isinstance(error, WindowClosedException)
        interrupted = isinstance(error, KeyboardInterrupt)
        if not window_closed and not interrupted:
            print_exc()
        # Close internal states
        self._close_pbar()
        self._close_chronometer()

    def _handle_completion(self):
        # Close internal states
        self._close_pbar()
        self._close_chronometer()

    def __call__(self) -> Annotation:
        """Stream audio chunks from `source` to `pipeline`.

        Returns
        -------
        predictions: Annotation
            Speaker diarization pipeline predictions
        """
        if self.show_progress:
            self._pbar.start()
        config = self.pipeline.config
        observable = self.stream
        if self.do_plot:
            # Buffering is needed for the real-time plot, so we do this at the very end
            observable = self.stream.pipe(
                dops.buffer_output(
                    duration=config.duration,
                    step=config.step,
                    latency=config.latency,
                    sample_rate=config.sample_rate,
                ),
                ops.do(StreamingPlot(config.duration, config.latency)),
            )
        observable.subscribe(
            on_error=self._handle_error,
            on_completed=self._handle_completion,
        )
        # FIXME if read() isn't blocking, the prediction returned is empty
        self.source.read()
        return self.accumulator.get_prediction()


class Benchmark:
    """
    Run an online speaker diarization pipeline on a set of audio files in batches.
    Write predictions to a given output directory.

    If the reference is given, calculate the average diarization error rate.

    Parameters
    ----------
    speech_path: Text or Path
        Directory with audio files.
    reference_path: Text, Path or None
        Directory with reference RTTM files (same names as audio files).
        If None, performance will not be calculated.
        Defaults to None.
    output_path: Text, Path or None
        Output directory to store predictions in RTTM format.
        If None, predictions will not be written to disk.
        Defaults to None.
    show_progress: bool
        Whether to show progress bars.
        Defaults to True.
    show_report: bool
        Whether to print a performance report to stdout.
        Defaults to True.
    batch_size: int
        Inference batch size.
        If < 2, then it will run in real time.
        If >= 2, then it will pre-calculate segmentation and
        embeddings, running the rest in real time.
        The performance between this two modes does not differ.
        Defaults to 32.
    """

    def __init__(
        self,
        speech_path: Union[Text, Path],
        reference_path: Optional[Union[Text, Path]] = None,
        output_path: Optional[Union[Text, Path]] = None,
        show_progress: bool = True,
        show_report: bool = True,
        batch_size: int = 32,
    ):
        self.speech_path = Path(speech_path).expanduser()
        assert self.speech_path.is_dir(), "Speech path must be a directory"

        # If there's no reference and no output, then benchmark has no output
        msg = "Benchmark expected reference path, output path or both"
        assert reference_path is not None or output_path is not None, msg

        self.reference_path = reference_path
        if reference_path is not None:
            self.reference_path = Path(self.reference_path).expanduser()
            assert self.reference_path.is_dir(), "Reference path must be a directory"

        self.output_path = output_path
        if self.output_path is not None:
            self.output_path = Path(output_path).expanduser()
            self.output_path.mkdir(parents=True, exist_ok=True)

        self.show_progress = show_progress
        self.show_report = show_report
        self.batch_size = batch_size

    def get_file_paths(self) -> List[Path]:
        """Return the path for each file in the benchmark.

        Returns
        -------
        paths: List[Path]
            List of audio file paths.
        """
        return list(self.speech_path.iterdir())

    def run_single(
        self,
        pipeline: blocks.Pipeline,
        filepath: Path,
        progress_bar: ProgressBar,
    ) -> Annotation:
        """Run a given pipeline on a given file.
        Note that this method does NOT reset the
        state of the pipeline before execution.

        Parameters
        ----------
        pipeline: StreamingPipeline
            Speaker diarization pipeline to run.
        filepath: Path
            Path to the target file.
        progress_bar: diart.progress.ProgressBar
            An object to manage the progress of this run.

        Returns
        -------
        prediction: Annotation
            Pipeline prediction for the given file.
        """
        padding = pipeline.config.get_file_padding(filepath)
        source = src.FileAudioSource(
            filepath,
            pipeline.config.sample_rate,
            padding,
            pipeline.config.step,
        )
        pipeline.set_timestamp_shift(-padding[0])
        inference = StreamingInference(
            pipeline,
            source,
            self.batch_size,
            do_profile=False,
            do_plot=False,
            show_progress=self.show_progress,
            progress_bar=progress_bar,
        )

        pred = inference()
        pred.uri = source.uri

        if self.output_path is not None:
            with open(self.output_path / f"{source.uri}.rttm", "w") as out_file:
                pred.write_rttm(out_file)

        return pred

    def evaluate(
        self,
        predictions: List[Annotation],
        metric: BaseMetric,
    ) -> Union[pd.DataFrame, List[Annotation]]:
        """If a reference path was provided,
        compute the diarization error rate of a list of predictions.

        Parameters
        ----------
        predictions: List[Annotation]
            Predictions to evaluate.
        metric: BaseMetric
            Evaluation metric from pyannote.metrics.

        Returns
        -------
        report_or_predictions: Union[pd.DataFrame, List[Annotation]]
            A performance report as a pandas `DataFrame` if a
            reference path was given. Otherwise return the same predictions.
        """
        if self.reference_path is not None:
            progress_bar = TQDMProgressBar(f"Computing {metric.name}", leave=False)
            progress_bar.create(total=len(predictions), unit="file")
            progress_bar.start()
            for hyp in predictions:
                ref = load_rttm(self.reference_path / f"{hyp.uri}.rttm").popitem()[1]
                metric(ref, hyp)
                progress_bar.update()
            progress_bar.close()
            return metric.report(display=self.show_report)
        return predictions

    def __call__(
        self,
        pipeline_class: type,
        config: blocks.PipelineConfig,
        metric: Optional[BaseMetric] = None,
    ) -> Union[pd.DataFrame, List[Annotation]]:
        """Run a given pipeline on a set of audio files.
        The internal state of the pipeline is reset before benchmarking.

        Parameters
        ----------
        pipeline_class: class
            Class from the StreamingPipeline hierarchy.
            A pipeline from this class will be instantiated by each worker.
        config: StreamingConfig
            Streaming pipeline configuration.
        metric: Optional[BaseMetric]
            Evaluation metric from pyannote.metrics.
            Defaults to the pipeline's suggested metric (see `StreamingPipeline.suggest_metric()`)

        Returns
        -------
        performance: pandas.DataFrame or List[Annotation]
            If reference annotations are given, a DataFrame with detailed
            performance on each file as well as average performance.

            If no reference annotations, a list of predictions.
        """
        audio_file_paths = self.get_file_paths()
        num_audio_files = len(audio_file_paths)
        pipeline = pipeline_class(config)

        predictions = []
        for i, filepath in enumerate(audio_file_paths):
            pipeline.reset()
            desc = f"Streaming {filepath.stem} ({i + 1}/{num_audio_files})"
            progress = TQDMProgressBar(desc, leave=False, do_close=True)
            predictions.append(self.run_single(pipeline, filepath, progress))

        metric = pipeline.suggest_metric() if metric is None else metric
        return self.evaluate(predictions, metric)


class Parallelize:
    """Wrapper to parallelize the execution of a `Benchmark` instance.
    Note that models will be copied in each worker instead of being reused.

    Parameters
    ----------
    benchmark: Benchmark
        Benchmark instance to execute in parallel.
    num_workers: int
        Number of parallel workers.
        Defaults to 0 (no parallelism).
    """

    def __init__(
        self,
        benchmark: Benchmark,
        num_workers: int = 4,
    ):
        self.benchmark = benchmark
        self.num_workers = num_workers

    def run_single_job(
        self,
        pipeline_class: type,
        config: blocks.PipelineConfig,
        filepath: Path,
        description: Text,
    ) -> Annotation:
        """Build and run a pipeline on a single file.
        Configure execution to show progress alongside parallel runs.

        Parameters
        ----------
        pipeline_class: class
            Class from the StreamingPipeline hierarchy.
            A pipeline from this class will be instantiated.
        config: StreamingConfig
            Streaming pipeline configuration.
        filepath: Path
            Path to the target file.
        description: Text
            Description to show in the parallel progress bar.

        Returns
        -------
        prediction: Annotation
            Pipeline prediction for the given file.
        """
        # The process ID inside the pool determines the position of the progress bar
        idx_process = int(current_process().name.split("-")[1]) - 1
        # TODO share models across processes
        # Instantiate a pipeline with the config
        pipeline = pipeline_class(config)
        # Create the progress bar for this job
        progress = TQDMProgressBar(
            description, leave=False, position=idx_process, do_close=True
        )
        # Run the pipeline
        return self.benchmark.run_single(pipeline, filepath, progress)

    def __call__(
        self,
        pipeline_class: type,
        config: blocks.PipelineConfig,
        metric: Optional[BaseMetric] = None,
    ) -> Union[pd.DataFrame, List[Annotation]]:
        """Run a given pipeline on a set of audio files in parallel.
        Each worker will build and run the pipeline on a different file.

        Parameters
        ----------
        pipeline_class: class
            Class from the StreamingPipeline hierarchy.
            A pipeline from this class will be instantiated by each worker.
        config: StreamingConfig
            Streaming pipeline configuration.
        metric: Optional[BaseMetric]
            Evaluation metric from pyannote.metrics.
            Defaults to the pipeline's suggested metric (see `StreamingPipeline.suggest_metric()`)

        Returns
        -------
        performance: pandas.DataFrame or List[Annotation]
            If reference annotations are given, a DataFrame with detailed
            performance on each file as well as average performance.

            If no reference annotations, a list of predictions.
        """
        audio_file_paths = self.benchmark.get_file_paths()
        num_audio_files = len(audio_file_paths)

        # Workaround for multiprocessing with GPU
        torch.multiprocessing.set_start_method("spawn")
        # For Windows support
        freeze_support()

        # Create the pool of workers using a lock for parallel tqdm usage
        pool = Pool(
            processes=self.num_workers, initargs=(RLock(),), initializer=tqdm.set_lock
        )
        # Determine the arguments for each job
        arg_list = [
            (
                pipeline_class,
                config,
                filepath,
                f"Streaming {filepath.stem} ({i + 1}/{num_audio_files})",
            )
            for i, filepath in enumerate(audio_file_paths)
        ]
        # Submit all jobs
        jobs = [pool.apply_async(self.run_single_job, args=args) for args in arg_list]

        # Wait and collect results
        pool.close()
        predictions = [job.get() for job in jobs]

        # Evaluate results
        metric = pipeline_class.suggest_metric() if metric is None else metric
        return self.benchmark.evaluate(predictions, metric)



from __future__ import annotations

from typing import Callable, Iterable, List, Optional, Text, Tuple, Union, Dict
from abc import ABC, abstractmethod

import numpy as np
from pyannote.core.utils.distance import cdist
from scipy.optimize import linear_sum_assignment as lsap


class MappingMatrixObjective(ABC):
    def invalid_tensor(self, shape: Union[Tuple, int]) -> np.ndarray:
        return np.ones(shape) * self.invalid_value

    def optimal_assignments(self, matrix: np.ndarray) -> List[int]:
        return list(lsap(matrix, self.maximize)[1])

    def mapped_indices(self, matrix: np.ndarray, axis: int) -> List[int]:
        # Entries full of invalid_value are not mapped
        best_values = self.best_value_fn(matrix, axis=axis)
        return list(np.where(best_values != self.invalid_value)[0])

    def hard_speaker_map(
        self, num_src: int, num_tgt: int, assignments: Iterable[Tuple[int, int]]
    ) -> SpeakerMap:
        """Create a hard map object where the highest cost is put
        everywhere except on hard assignments from ``assignments``.

        Parameters
        ----------
        num_src: int
            Number of source speakers
        num_tgt: int
            Number of target speakers
        assignments: Iterable[Tuple[int, int]]
            An iterable of tuples with two elements having the first element as the source speaker
            and the second element as the target speaker

        Returns
        -------
            SpeakerMap
        """
        mapping_matrix = self.invalid_tensor(shape=(num_src, num_tgt))
        for src, tgt in assignments:
            mapping_matrix[src, tgt] = self.best_possible_value
        return SpeakerMap(mapping_matrix, self)

    @property
    def invalid_value(self) -> float:
        # linear_sum_assignment cannot deal with np.inf,
        # which would be ideal. Using a big number instead.
        return -1e10 if self.maximize else 1e10

    @property
    @abstractmethod
    def maximize(self) -> bool:
        pass

    @property
    @abstractmethod
    def best_possible_value(self) -> float:
        pass

    @property
    @abstractmethod
    def best_value_fn(self) -> Callable:
        pass


class MinimizationObjective(MappingMatrixObjective):
    @property
    def maximize(self) -> bool:
        return False

    @property
    def best_possible_value(self) -> float:
        return 0

    @property
    def best_value_fn(self) -> Callable:
        return np.min


class MaximizationObjective(MappingMatrixObjective):
    def __init__(self, max_value: float = 1):
        self.max_value = max_value

    @property
    def maximize(self) -> bool:
        return True

    @property
    def best_possible_value(self) -> float:
        return self.max_value

    @property
    def best_value_fn(self) -> Callable:
        return np.max


class SpeakerMapBuilder:
    @staticmethod
    def hard_map(
        shape: Tuple[int, int], assignments: Iterable[Tuple[int, int]], maximize: bool
    ) -> SpeakerMap:
        """Create a ``SpeakerMap`` object based on the given assignments. This is a "hard" map, meaning that the
        highest cost is put everywhere except on hard assignments from ``assignments``.

        Parameters
        ----------
        shape: Tuple[int, int])
            Shape of the mapping matrix
        assignments: Iterable[Tuple[int, int]]
            An iterable of tuples with two elements having the first element as the source speaker
            and the second element as the target speaker
        maximize: bool
            whether to use scores where higher is better (true) or where lower is better (false)

        Returns
        -------
        SpeakerMap
        """
        num_src, num_tgt = shape
        objective = MaximizationObjective if maximize else MinimizationObjective
        return objective().hard_speaker_map(num_src, num_tgt, assignments)

    @staticmethod
    def correlation(scores1: np.ndarray, scores2: np.ndarray) -> SpeakerMap:
        score_matrix_per_frame = (
            np.stack(  # (local_speakers, num_frames, global_speakers)
                [
                    scores1[:, speaker : speaker + 1] * scores2
                    for speaker in range(scores1.shape[1])
                ],
                axis=0,
            )
        )
        # Calculate total speech "activations" per local speaker
        local_speech_scores = np.sum(scores1, axis=0).reshape(-1, 1)
        # Calculate speaker mapping matrix
        # Cost matrix is the correlation divided by sum of local activations
        score_matrix = np.sum(score_matrix_per_frame, axis=1) / local_speech_scores
        # We want to maximize the correlation to calculate optimal speaker alignments
        return SpeakerMap(score_matrix, MaximizationObjective(max_value=1))

    @staticmethod
    def mse(scores1: np.ndarray, scores2: np.ndarray) -> SpeakerMap:
        cost_matrix = np.stack(  # (local_speakers, local_speakers)
            [
                np.square(scores1[:, speaker : speaker + 1] - scores2).mean(axis=0)
                for speaker in range(scores1.shape[1])
            ],
            axis=0,
        )
        # We want to minimize the MSE to calculate optimal speaker alignments
        return SpeakerMap(cost_matrix, MinimizationObjective())

    @staticmethod
    def mae(scores1: np.ndarray, scores2: np.ndarray) -> SpeakerMap:
        cost_matrix = np.stack(  # (local_speakers, local_speakers)
            [
                np.abs(scores1[:, speaker : speaker + 1] - scores2).mean(axis=0)
                for speaker in range(scores1.shape[1])
            ],
            axis=0,
        )
        # We want to minimize the MSE to calculate optimal speaker alignments
        return SpeakerMap(cost_matrix, MinimizationObjective())

    @staticmethod
    def dist(
        embeddings1: np.ndarray, embeddings2: np.ndarray, metric: Text = "cosine"
    ) -> SpeakerMap:
        # We want to minimize the distance to calculate optimal speaker alignments
        dist_matrix = cdist(embeddings1, embeddings2, metric=metric)
        return SpeakerMap(dist_matrix, MinimizationObjective())


class SpeakerMap:
    def __init__(self, mapping_matrix: np.ndarray, objective: MappingMatrixObjective):
        self.mapping_matrix = mapping_matrix
        self.objective = objective
        self.num_source_speakers = self.mapping_matrix.shape[0]
        self.num_target_speakers = self.mapping_matrix.shape[1]
        self.mapped_source_speakers = self.objective.mapped_indices(
            self.mapping_matrix, axis=1
        )
        self.mapped_target_speakers = self.objective.mapped_indices(
            self.mapping_matrix, axis=0
        )
        self._opt_assignments: Optional[List[int]] = None

    @property
    def _raw_optimal_assignments(self) -> List[int]:
        if self._opt_assignments is None:
            self._opt_assignments = self.objective.optimal_assignments(
                self.mapping_matrix
            )
        return self._opt_assignments

    @property
    def shape(self) -> Tuple[int, int]:
        return self.mapping_matrix.shape

    def __len__(self):
        return len(self.mapped_source_speakers)

    def __add__(self, other: SpeakerMap) -> SpeakerMap:
        return self.union(other)

    def _strict_check_valid(self, src: int, tgt: int) -> bool:
        return self.mapping_matrix[src, tgt] != self.objective.invalid_value

    def _loose_check_valid(self, src: int, tgt: int) -> bool:
        return self.is_source_speaker_mapped(src)

    def valid_assignments(
        self,
        strict: bool = False,
        as_array: bool = False,
    ) -> Union[Tuple[List[int], List[int]], Tuple[np.ndarray, np.ndarray]]:
        source, target = [], []
        val_type = "strict" if strict else "loose"
        is_valid = getattr(self, f"_{val_type}_check_valid")
        for src, tgt in enumerate(self._raw_optimal_assignments):
            if is_valid(src, tgt):
                source.append(src)
                target.append(tgt)
        if as_array:
            source, target = np.array(source), np.array(target)
        return source, target

    def to_dict(self, strict: bool = False) -> Dict[int, int]:
        return {src: tgt for src, tgt in zip(*self.valid_assignments(strict))}

    def to_inverse_dict(self, strict: bool = False) -> Dict[int, int]:
        return {tgt: src for src, tgt in zip(*self.valid_assignments(strict))}

    def is_source_speaker_mapped(self, source_speaker: int) -> bool:
        return source_speaker in self.mapped_source_speakers

    def is_target_speaker_mapped(self, target_speaker: int) -> bool:
        return target_speaker in self.mapped_target_speakers

    def set_source_speaker(self, src_speaker, tgt_speaker: int):
        # if not force:
        #     assert not self.is_source_speaker_mapped(src_speaker)
        #     assert not self.is_target_speaker_mapped(tgt_speaker)
        new_cost_matrix = np.copy(self.mapping_matrix)
        new_cost_matrix[src_speaker, tgt_speaker] = self.objective.best_possible_value
        return SpeakerMap(new_cost_matrix, self.objective)

    def unmap_source_speaker(self, src_speaker: int):
        new_cost_matrix = np.copy(self.mapping_matrix)
        new_cost_matrix[src_speaker] = self.objective.invalid_tensor(
            shape=self.num_target_speakers
        )
        return SpeakerMap(new_cost_matrix, self.objective)

    def unmap_threshold(self, threshold: float) -> SpeakerMap:
        def is_invalid(val):
            if self.objective.maximize:
                return val <= threshold
            else:
                return val >= threshold

        return self.unmap_speakers(
            [
                src
                for src, tgt in zip(*self.valid_assignments())
                if is_invalid(self.mapping_matrix[src, tgt])
            ]
        )

    def unmap_speakers(
        self,
        source_speakers: Optional[Union[List[int], np.ndarray]] = None,
        target_speakers: Optional[Union[List[int], np.ndarray]] = None,
    ) -> SpeakerMap:
        # Set invalid_value to disabled speakers.
        # If they happen to be the best mapping for a local speaker,
        # it means that the mapping of the local speaker should be ignored.
        source_speakers = [] if source_speakers is None else source_speakers
        target_speakers = [] if target_speakers is None else target_speakers
        new_cost_matrix = np.copy(self.mapping_matrix)
        for speaker1 in source_speakers:
            new_cost_matrix[speaker1] = self.objective.invalid_tensor(
                shape=self.num_target_speakers
            )
        for speaker2 in target_speakers:
            new_cost_matrix[:, speaker2] = self.objective.invalid_tensor(
                shape=self.num_source_speakers
            )
        return SpeakerMap(new_cost_matrix, self.objective)

    def compose(self, other: SpeakerMap) -> SpeakerMap:
        """Let's say that `self` is a mapping of `source_speakers` to `intermediate_speakers`
        and `other` is a mapping from `intermediate_speakers` to `target_speakers`.

        Compose `self` with `other` to obtain a new mapping from `source_speakers` to `target_speakers`.
        """
        new_cost_matrix = other.objective.invalid_tensor(
            shape=(self.num_source_speakers, other.num_target_speakers)
        )
        for src_speaker, intermediate_speaker in zip(*self.valid_assignments()):
            target_speaker = other.mapping_matrix[intermediate_speaker]
            new_cost_matrix[src_speaker] = target_speaker
        return SpeakerMap(new_cost_matrix, other.objective)

    def union(self, other: SpeakerMap):
        """`self` and `other` are two maps with the same dimensions.
        Return a new hard speaker map containing assignments in both maps.

        An assignment from `other` is ignored if it is in conflict with
        a source or target speaker from `self`.

        WARNING: The resulting map doesn't preserve soft assignments
        because `self` and `other` might have different objectives.

        :param other: SpeakerMap
            Another speaker map
        """
        assert self.shape == other.shape
        best_val = self.objective.best_possible_value
        new_cost_matrix = self.objective.invalid_tensor(self.shape)
        self_src, self_tgt = self.valid_assignments()
        other_src, other_tgt = other.valid_assignments()
        for src in range(self.num_source_speakers):
            if src in self_src:
                # `self` is preserved by default
                tgt = self_tgt[self_src.index(src)]
                new_cost_matrix[src, tgt] = best_val
            elif src in other_src:
                # In order to add an assignment from `other`,
                # the target speaker cannot be in conflict with `self`
                tgt = other_tgt[other_src.index(src)]
                if not self.is_target_speaker_mapped(tgt):
                    new_cost_matrix[src, tgt] = best_val
        return SpeakerMap(new_cost_matrix, self.objective)

    def apply(self, source_scores: np.ndarray) -> np.ndarray:
        """Apply this mapping to a score matrix of source speakers
        to obtain the same scores aligned to target speakers.

        Parameters
        ----------
        source_scores : SlidingWindowFeature, (num_frames, num_source_speakers)
            Source speaker scores per frame.

        Returns
        -------
        projected_scores : SlidingWindowFeature, (num_frames, num_target_speakers)
            Score matrix for target speakers.
        """
        # Map local speaker scores to the most probable global speaker. Unknown scores are set to 0
        num_frames = source_scores.data.shape[0]
        projected_scores = np.zeros((num_frames, self.num_target_speakers))
        for src_speaker, tgt_speaker in zip(*self.valid_assignments()):
            projected_scores[:, tgt_speaker] = source_scores[:, src_speaker]
        return projected_scores



from pathlib import Path
from typing import Text, Union

import torch
import torchaudio
from torchaudio.functional import resample

torchaudio.set_audio_backend("soundfile")


FilePath = Union[Text, Path]


class AudioLoader:
    def __init__(self, sample_rate: int, mono: bool = True):
        self.sample_rate = sample_rate
        self.mono = mono

    def load(self, filepath: FilePath) -> torch.Tensor:
        """Load an audio file into a torch.Tensor.

        Parameters
        ----------
        filepath : FilePath
            Path to an audio file

        Returns
        -------
        waveform : torch.Tensor, shape (channels, samples)
        """
        waveform, sample_rate = torchaudio.load(filepath)
        # Get channel mean if mono
        if self.mono and waveform.shape[0] > 1:
            waveform = waveform.mean(dim=0, keepdim=True)
        # Resample if needed
        if self.sample_rate != sample_rate:
            waveform = resample(waveform, sample_rate, self.sample_rate)
        return waveform

    @staticmethod
    def get_duration(filepath: FilePath) -> float:
        """Get audio file duration in seconds.

        Parameters
        ----------
        filepath : FilePath
            Path to an audio file.

        Returns
        -------
        duration : float
            Duration in seconds.
        """
        info = torchaudio.info(filepath)
        return info.num_frames / info.sample_rate



from abc import ABC, abstractmethod
from pathlib import Path
from queue import SimpleQueue
from typing import Text, Optional, AnyStr, Dict, Any, Union, Tuple

import numpy as np
import sounddevice as sd
import torch
from einops import rearrange
from rx.subject import Subject
from torchaudio.io import StreamReader
from websocket_server import WebsocketServer

from . import utils
from .audio import FilePath, AudioLoader


class AudioSource(ABC):
    """Represents a source of audio that can start streaming via the `stream` property.

    Parameters
    ----------
    uri: Text
        Unique identifier of the audio source.
    sample_rate: int
        Sample rate of the audio source.
    """

    def __init__(self, uri: Text, sample_rate: int):
        self.uri = uri
        self.sample_rate = sample_rate
        self.stream = Subject()

    @property
    def duration(self) -> Optional[float]:
        """The duration of the stream if known. Defaults to None (unknown duration)."""
        return None

    @abstractmethod
    def read(self):
        """Start reading the source and yielding samples through the stream."""
        pass

    @abstractmethod
    def close(self):
        """Stop reading the source and close all open streams."""
        pass


class FileAudioSource(AudioSource):
    """Represents an audio source tied to a file.

    Parameters
    ----------
    file: FilePath
        Path to the file to stream.
    sample_rate: int
        Sample rate of the chunks emitted.
    padding: (float, float)
        Left and right padding to add to the file (in seconds).
        Defaults to (0, 0).
    block_duration: int
        Duration of each emitted chunk in seconds.
        Defaults to 0.5 seconds.
    """

    def __init__(
        self,
        file: FilePath,
        sample_rate: int,
        padding: Tuple[float, float] = (0, 0),
        block_duration: float = 0.5,
    ):
        super().__init__(Path(file).stem, sample_rate)
        self.loader = AudioLoader(self.sample_rate, mono=True)
        self._duration = self.loader.get_duration(file)
        self.file = file
        self.resolution = 1 / self.sample_rate
        self.block_size = int(np.rint(block_duration * self.sample_rate))
        self.padding_start, self.padding_end = padding
        self.is_closed = False

    @property
    def duration(self) -> Optional[float]:
        # The duration of a file is known
        return self.padding_start + self._duration + self.padding_end

    def read(self):
        """Send each chunk of samples through the stream"""
        waveform = self.loader.load(self.file)

        # Add zero padding at the beginning if required
        if self.padding_start > 0:
            num_pad_samples = int(np.rint(self.padding_start * self.sample_rate))
            zero_padding = torch.zeros(waveform.shape[0], num_pad_samples)
            waveform = torch.cat([zero_padding, waveform], dim=1)

        # Add zero padding at the end if required
        if self.padding_end > 0:
            num_pad_samples = int(np.rint(self.padding_end * self.sample_rate))
            zero_padding = torch.zeros(waveform.shape[0], num_pad_samples)
            waveform = torch.cat([waveform, zero_padding], dim=1)

        # Split into blocks
        _, num_samples = waveform.shape
        chunks = rearrange(
            waveform.unfold(1, self.block_size, self.block_size),
            "channel chunk sample -> chunk channel sample",
        ).numpy()

        # Add last incomplete chunk with padding
        if num_samples % self.block_size != 0:
            last_chunk = (
                waveform[:, chunks.shape[0] * self.block_size :].unsqueeze(0).numpy()
            )
            diff_samples = self.block_size - last_chunk.shape[-1]
            last_chunk = np.concatenate(
                [last_chunk, np.zeros((1, 1, diff_samples))], axis=-1
            )
            chunks = np.vstack([chunks, last_chunk])

        # Stream blocks
        for i, waveform in enumerate(chunks):
            try:
                if self.is_closed:
                    break
                self.stream.on_next(waveform)
            except BaseException as e:
                self.stream.on_error(e)
                break
        self.stream.on_completed()
        self.close()

    def close(self):
        self.is_closed = True


class MicrophoneAudioSource(AudioSource):
    """Audio source tied to a local microphone.

    Parameters
    ----------
    block_duration: int
        Duration of each emitted chunk in seconds.
        Defaults to 0.5 seconds.
    device: int | str | (int, str) | None
        Device identifier compatible for the sounddevice stream.
        If None, use the default device.
        Defaults to None.
    """

    def __init__(
        self,
        block_duration: float = 0.5,
        device: Optional[Union[int, Text, Tuple[int, Text]]] = None,
    ):
        # Use the lowest supported sample rate
        sample_rates = [16000, 32000, 44100, 48000]
        best_sample_rate = None
        for sr in sample_rates:
            try:
                sd.check_input_settings(device=device, samplerate=sr)
            except Exception:
                pass
            else:
                best_sample_rate = sr
                break
        super().__init__(f"input_device:{device}", best_sample_rate)

        # Determine block size in samples and create input stream
        self.block_size = int(np.rint(block_duration * self.sample_rate))
        self._mic_stream = sd.InputStream(
            channels=1,
            samplerate=self.sample_rate,
            latency=0,
            blocksize=self.block_size,
            callback=self._read_callback,
            device=device,
        )
        self._queue = SimpleQueue()

    def _read_callback(self, samples, *args):
        self._queue.put_nowait(samples[:, [0]].T)

    def read(self):
        self._mic_stream.start()
        while self._mic_stream:
            try:
                while self._queue.empty():
                    if self._mic_stream.closed:
                        break
                self.stream.on_next(self._queue.get_nowait())
            except BaseException as e:
                self.stream.on_error(e)
                break
        self.stream.on_completed()
        self.close()

    def close(self):
        self._mic_stream.stop()
        self._mic_stream.close()


class WebSocketAudioSource(AudioSource):
    """Represents a source of audio coming from the network using the WebSocket protocol.

    Parameters
    ----------
    sample_rate: int
        Sample rate of the chunks emitted.
    host: Text
        The host to run the websocket server.
        Defaults to 127.0.0.1.
    port: int
        The port to run the websocket server.
        Defaults to 7007.
    key: Text | Path | None
        Path to a key if using SSL.
        Defaults to no key.
    certificate: Text | Path | None
        Path to a certificate if using SSL.
        Defaults to no certificate.
    """

    def __init__(
        self,
        sample_rate: int,
        host: Text = "127.0.0.1",
        port: int = 7007,
        key: Optional[Union[Text, Path]] = None,
        certificate: Optional[Union[Text, Path]] = None,
    ):
        # FIXME sample_rate is not being used, this can be confusing and lead to incompatibilities.
        #  I would prefer the client to send a JSON with data and sample rate, then resample if needed
        super().__init__(f"{host}:{port}", sample_rate)
        self.client: Optional[Dict[Text, Any]] = None
        self.server = WebsocketServer(host, port, key=key, cert=certificate)
        self.server.set_fn_message_received(self._on_message_received)

    def _on_message_received(
        self,
        client: Dict[Text, Any],
        server: WebsocketServer,
        message: AnyStr,
    ):
        # Only one client at a time is allowed
        if self.client is None or self.client["id"] != client["id"]:
            self.client = client
        # Send decoded audio to pipeline
        self.stream.on_next(utils.decode_audio(message))

    def read(self):
        """Starts running the websocket server and listening for audio chunks"""
        self.server.run_forever()

    def close(self):
        """Close the websocket server"""
        if self.server is not None:
            self.stream.on_completed()
            self.server.shutdown_gracefully()

    def send(self, message: AnyStr):
        """Send a message through the current websocket.

        Parameters
        ----------
        message: AnyStr
            Bytes or string to send.
        """
        if len(message) > 0:
            self.server.send_message(self.client, message)


class TorchStreamAudioSource(AudioSource):
    def __init__(
        self,
        uri: Text,
        sample_rate: int,
        streamer: StreamReader,
        stream_index: Optional[int] = None,
        block_duration: float = 0.5,
    ):
        super().__init__(uri, sample_rate)
        self.block_size = int(np.rint(block_duration * self.sample_rate))
        self._streamer = streamer
        self._streamer.add_basic_audio_stream(
            frames_per_chunk=self.block_size,
            stream_index=stream_index,
            format="fltp",
            sample_rate=self.sample_rate,
        )
        self.is_closed = False

    def read(self):
        for item in self._streamer.stream():
            try:
                if self.is_closed:
                    break
                # shape (samples, channels) to (1, samples)
                chunk = np.mean(item[0].numpy(), axis=1, keepdims=True).T
                self.stream.on_next(chunk)
            except BaseException as e:
                self.stream.on_error(e)
                break
        self.stream.on_completed()
        self.close()

    def close(self):
        self.is_closed = True


class AppleDeviceAudioSource(TorchStreamAudioSource):
    def __init__(
        self,
        sample_rate: int,
        device: str = "0:0",
        stream_index: int = 0,
        block_duration: float = 0.5,
    ):
        uri = f"apple_input_device:{device}"
        streamer = StreamReader(device, format="avfoundation")
        super().__init__(uri, sample_rate, streamer, stream_index, block_duration)



import argparse
from pathlib import Path

import torch

from diart import argdoc
from diart import models as m
from diart import sources as src
from diart import utils
from diart.inference import StreamingInference
from diart.sinks import RTTMWriter


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument("--host", default="0.0.0.0", type=str, help="Server host")
    parser.add_argument("--port", default=7007, type=int, help="Server port")
    parser.add_argument(
        "--pipeline",
        default="SpeakerDiarization",
        type=str,
        help="Class of the pipeline to optimize. Defaults to 'SpeakerDiarization'",
    )
    parser.add_argument(
        "--segmentation",
        default="pyannote/segmentation",
        type=str,
        help=f"{argdoc.SEGMENTATION}. Defaults to pyannote/segmentation",
    )
    parser.add_argument(
        "--embedding",
        default="pyannote/embedding",
        type=str,
        help=f"{argdoc.EMBEDDING}. Defaults to pyannote/embedding",
    )
    parser.add_argument(
        "--duration",
        type=float,
        help=f"{argdoc.DURATION}. Defaults to training segmentation duration",
    )
    parser.add_argument(
        "--step", default=0.5, type=float, help=f"{argdoc.STEP}. Defaults to 0.5"
    )
    parser.add_argument(
        "--latency", default=0.5, type=float, help=f"{argdoc.LATENCY}. Defaults to 0.5"
    )
    parser.add_argument(
        "--tau-active", default=0.5, type=float, help=f"{argdoc.TAU}. Defaults to 0.5"
    )
    parser.add_argument(
        "--rho-update", default=0.3, type=float, help=f"{argdoc.RHO}. Defaults to 0.3"
    )
    parser.add_argument(
        "--delta-new", default=1, type=float, help=f"{argdoc.DELTA}. Defaults to 1"
    )
    parser.add_argument(
        "--gamma", default=3, type=float, help=f"{argdoc.GAMMA}. Defaults to 3"
    )
    parser.add_argument(
        "--beta", default=10, type=float, help=f"{argdoc.BETA}. Defaults to 10"
    )
    parser.add_argument(
        "--max-speakers",
        default=20,
        type=int,
        help=f"{argdoc.MAX_SPEAKERS}. Defaults to 20",
    )
    parser.add_argument(
        "--cpu",
        dest="cpu",
        action="store_true",
        help=f"{argdoc.CPU}. Defaults to GPU if available, CPU otherwise",
    )
    parser.add_argument(
        "--output", type=Path, help=f"{argdoc.OUTPUT}. Defaults to no writing"
    )
    parser.add_argument(
        "--hf-token",
        default="true",
        type=str,
        help=f"{argdoc.HF_TOKEN}. Defaults to 'true' (required by pyannote)",
    )
    args = parser.parse_args()

    # Resolve device
    args.device = torch.device("cpu") if args.cpu else None

    # Resolve models
    hf_token = utils.parse_hf_token_arg(args.hf_token)
    args.segmentation = m.SegmentationModel.from_pyannote(args.segmentation, hf_token)
    args.embedding = m.EmbeddingModel.from_pyannote(args.embedding, hf_token)

    # Resolve pipeline
    pipeline_class = utils.get_pipeline_class(args.pipeline)
    config = pipeline_class.get_config_class()(**vars(args))
    pipeline = pipeline_class(config)

    # Create websocket audio source
    audio_source = src.WebSocketAudioSource(config.sample_rate, args.host, args.port)

    # Run online inference
    inference = StreamingInference(
        pipeline,
        audio_source,
        batch_size=1,
        do_profile=False,
        do_plot=False,
        show_progress=True,
    )

    # Write to disk if required
    if args.output is not None:
        inference.attach_observers(
            RTTMWriter(audio_source.uri, args.output / f"{audio_source.uri}.rttm")
        )

    # Send back responses as RTTM text lines
    inference.attach_hooks(lambda ann_wav: audio_source.send(ann_wav[0].to_rttm()))

    # Run server and pipeline
    inference()


if __name__ == "__main__":
    run()



import argparse
from pathlib import Path

import optuna
import torch
from optuna.samplers import TPESampler

from diart import argdoc
from diart import models as m
from diart import utils
from diart.blocks.base import HyperParameter
from diart.optim import Optimizer


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "root",
        type=str,
        help="Directory with audio files CONVERSATION.(wav|flac|m4a|...)",
    )
    parser.add_argument(
        "--reference",
        required=True,
        type=str,
        help="Directory with RTTM files CONVERSATION.rttm. Names must match audio files",
    )
    parser.add_argument(
        "--pipeline",
        default="SpeakerDiarization",
        type=str,
        help="Class of the pipeline to optimize. Defaults to 'SpeakerDiarization'",
    )
    parser.add_argument(
        "--segmentation",
        default="pyannote/segmentation",
        type=str,
        help=f"{argdoc.SEGMENTATION}. Defaults to pyannote/segmentation",
    )
    parser.add_argument(
        "--embedding",
        default="pyannote/embedding",
        type=str,
        help=f"{argdoc.EMBEDDING}. Defaults to pyannote/embedding",
    )
    parser.add_argument(
        "--duration",
        type=float,
        help=f"{argdoc.DURATION}. Defaults to training segmentation duration",
    )
    parser.add_argument(
        "--step", default=0.5, type=float, help=f"{argdoc.STEP}. Defaults to 0.5"
    )
    parser.add_argument(
        "--latency", default=0.5, type=float, help=f"{argdoc.LATENCY}. Defaults to 0.5"
    )
    parser.add_argument(
        "--tau-active", default=0.5, type=float, help=f"{argdoc.TAU}. Defaults to 0.5"
    )
    parser.add_argument(
        "--rho-update", default=0.3, type=float, help=f"{argdoc.RHO}. Defaults to 0.3"
    )
    parser.add_argument(
        "--delta-new", default=1, type=float, help=f"{argdoc.DELTA}. Defaults to 1"
    )
    parser.add_argument(
        "--gamma", default=3, type=float, help=f"{argdoc.GAMMA}. Defaults to 3"
    )
    parser.add_argument(
        "--beta", default=10, type=float, help=f"{argdoc.BETA}. Defaults to 10"
    )
    parser.add_argument(
        "--max-speakers",
        default=20,
        type=int,
        help=f"{argdoc.MAX_SPEAKERS}. Defaults to 20",
    )
    parser.add_argument(
        "--batch-size",
        default=32,
        type=int,
        help=f"{argdoc.BATCH_SIZE}. Defaults to 32",
    )
    parser.add_argument(
        "--cpu",
        dest="cpu",
        action="store_true",
        help=f"{argdoc.CPU}. Defaults to GPU if available, CPU otherwise",
    )
    parser.add_argument(
        "--hparams",
        nargs="+",
        default=("tau_active", "rho_update", "delta_new"),
        help="Hyper-parameters to optimize. Must match names in `PipelineConfig`. Defaults to tau_active, rho_update and delta_new",
    )
    parser.add_argument(
        "--num-iter", default=100, type=int, help="Number of optimization trials"
    )
    parser.add_argument(
        "--storage",
        type=str,
        help="Optuna storage string. If provided, continue a previous study instead of creating one. The database name must match the study name",
    )
    parser.add_argument("--output", type=str, help="Working directory")
    parser.add_argument(
        "--hf-token",
        default="true",
        type=str,
        help=f"{argdoc.HF_TOKEN}. Defaults to 'true' (required by pyannote)",
    )
    args = parser.parse_args()

    # Resolve device
    args.device = torch.device("cpu") if args.cpu else None

    # Resolve models
    hf_token = utils.parse_hf_token_arg(args.hf_token)
    args.segmentation = m.SegmentationModel.from_pyannote(args.segmentation, hf_token)
    args.embedding = m.EmbeddingModel.from_pyannote(args.embedding, hf_token)

    # Retrieve pipeline class
    pipeline_class = utils.get_pipeline_class(args.pipeline)

    # Create the base configuration for each trial
    base_config = pipeline_class.get_config_class()(**vars(args))

    # Create hyper-parameters to optimize
    possible_hparams = pipeline_class.hyper_parameters()
    hparams = [HyperParameter.from_name(name) for name in args.hparams]
    hparams = [hp for hp in hparams if hp in possible_hparams]
    if not hparams:
        print(
            f"No hyper-parameters to optimize. "
            f"Make sure to select one of: {', '.join([hp.name for hp in possible_hparams])}"
        )
        exit(1)

    # Use a custom storage if given
    if args.output is not None:
        msg = "Both `output` and `storage` were set, but only one was expected"
        assert args.storage is None, msg
        args.output = Path(args.output).expanduser()
        args.output.mkdir(parents=True, exist_ok=True)
        study_or_path = args.output
    elif args.storage is not None:
        db_name = Path(args.storage).stem
        study_or_path = optuna.load_study(db_name, args.storage, TPESampler())
    else:
        msg = "Please provide either `output` or `storage`"
        raise ValueError(msg)

    # Run optimization
    Optimizer(
        pipeline_class=pipeline_class,
        speech_path=args.root,
        reference_path=args.reference,
        study_or_path=study_or_path,
        batch_size=args.batch_size,
        hparams=hparams,
        base_config=base_config,
    )(num_iter=args.num_iter, show_progress=True)


if __name__ == "__main__":
    run()



from __future__ import annotations

from typing import Sequence

import numpy as np
import torch
from pyannote.core import Annotation, SlidingWindowFeature, SlidingWindow, Segment
from pyannote.metrics.base import BaseMetric
from pyannote.metrics.diarization import DiarizationErrorRate
from typing_extensions import Literal

from . import base
from .aggregation import DelayedAggregation
from .clustering import OnlineSpeakerClustering
from .embedding import OverlapAwareSpeakerEmbedding
from .segmentation import SpeakerSegmentation
from .utils import Binarize
from .. import models as m


class SpeakerDiarizationConfig(base.PipelineConfig):
    def __init__(
        self,
        segmentation: m.SegmentationModel | None = None,
        embedding: m.EmbeddingModel | None = None,
        duration: float | None = None,
        step: float = 0.5,
        latency: float | Literal["max", "min"] | None = None,
        tau_active: float = 0.6,
        rho_update: float = 0.3,
        delta_new: float = 1,
        gamma: float = 3,
        beta: float = 10,
        max_speakers: int = 20,
        device: torch.device | None = None,
        **kwargs,
    ):
        # Default segmentation model is pyannote/segmentation
        self.segmentation = segmentation or m.SegmentationModel.from_pyannote(
            "pyannote/segmentation"
        )

        # Default embedding model is pyannote/embedding
        self.embedding = embedding or m.EmbeddingModel.from_pyannote(
            "pyannote/embedding"
        )

        self._duration = duration
        self._sample_rate: int | None = None

        # Latency defaults to the step duration
        self._step = step
        self._latency = latency
        if self._latency is None or self._latency == "min":
            self._latency = self._step
        elif self._latency == "max":
            self._latency = self._duration

        self.tau_active = tau_active
        self.rho_update = rho_update
        self.delta_new = delta_new
        self.gamma = gamma
        self.beta = beta
        self.max_speakers = max_speakers

        self.device = device or torch.device(
            "cuda" if torch.cuda.is_available() else "cpu"
        )

    @property
    def duration(self) -> float:
        # Default duration is the one given by the segmentation model
        if self._duration is None:
            self._duration = self.segmentation.duration
        return self._duration

    @property
    def step(self) -> float:
        return self._step

    @property
    def latency(self) -> float:
        return self._latency

    @property
    def sample_rate(self) -> int:
        # Expected sample rate is given by the segmentation model
        if self._sample_rate is None:
            self._sample_rate = self.segmentation.sample_rate
        return self._sample_rate


class SpeakerDiarization(base.Pipeline):
    def __init__(self, config: SpeakerDiarizationConfig | None = None):
        self._config = SpeakerDiarizationConfig() if config is None else config

        msg = f"Latency should be in the range [{self._config.step}, {self._config.duration}]"
        assert self._config.step <= self._config.latency <= self._config.duration, msg

        self.segmentation = SpeakerSegmentation(
            self._config.segmentation, self._config.device
        )
        self.embedding = OverlapAwareSpeakerEmbedding(
            self._config.embedding,
            self._config.gamma,
            self._config.beta,
            norm=1,
            device=self._config.device,
        )
        self.pred_aggregation = DelayedAggregation(
            self._config.step,
            self._config.latency,
            strategy="hamming",
            cropping_mode="loose",
        )
        self.audio_aggregation = DelayedAggregation(
            self._config.step,
            self._config.latency,
            strategy="first",
            cropping_mode="center",
        )
        self.binarize = Binarize(self._config.tau_active)

        # Internal state, handle with care
        self.timestamp_shift = 0
        self.clustering = None
        self.chunk_buffer, self.pred_buffer = [], []
        self.reset()

    @staticmethod
    def get_config_class() -> type:
        return SpeakerDiarizationConfig

    @staticmethod
    def suggest_metric() -> BaseMetric:
        return DiarizationErrorRate(collar=0, skip_overlap=False)

    @staticmethod
    def hyper_parameters() -> Sequence[base.HyperParameter]:
        return [base.TauActive, base.RhoUpdate, base.DeltaNew]

    @property
    def config(self) -> SpeakerDiarizationConfig:
        return self._config

    def set_timestamp_shift(self, shift: float):
        self.timestamp_shift = shift

    def reset(self):
        self.set_timestamp_shift(0)
        self.clustering = OnlineSpeakerClustering(
            self.config.tau_active,
            self.config.rho_update,
            self.config.delta_new,
            "cosine",
            self.config.max_speakers,
        )
        self.chunk_buffer, self.pred_buffer = [], []

    def __call__(
        self, waveforms: Sequence[SlidingWindowFeature]
    ) -> Sequence[tuple[Annotation, SlidingWindowFeature]]:
        batch_size = len(waveforms)
        msg = "Pipeline expected at least 1 input"
        assert batch_size >= 1, msg

        # Create batch from chunk sequence, shape (batch, samples, channels)
        batch = torch.stack([torch.from_numpy(w.data) for w in waveforms])

        expected_num_samples = int(
            np.rint(self.config.duration * self.config.sample_rate)
        )
        msg = f"Expected {expected_num_samples} samples per chunk, but got {batch.shape[1]}"
        assert batch.shape[1] == expected_num_samples, msg

        # Extract segmentation and embeddings
        segmentations = self.segmentation(batch)  # shape (batch, frames, speakers)
        embeddings = self.embedding(
            batch, segmentations
        )  # shape (batch, speakers, emb_dim)

        seg_resolution = waveforms[0].extent.duration / segmentations.shape[1]

        outputs = []
        for wav, seg, emb in zip(waveforms, segmentations, embeddings):
            # Add timestamps to segmentation
            sw = SlidingWindow(
                start=wav.extent.start,
                duration=seg_resolution,
                step=seg_resolution,
            )
            seg = SlidingWindowFeature(seg.cpu().numpy(), sw)

            # Update clustering state and permute segmentation
            permuted_seg = self.clustering(seg, emb)

            # Update sliding buffer
            self.chunk_buffer.append(wav)
            self.pred_buffer.append(permuted_seg)

            # Aggregate buffer outputs for this time step
            agg_waveform = self.audio_aggregation(self.chunk_buffer)
            agg_prediction = self.pred_aggregation(self.pred_buffer)
            agg_prediction = self.binarize(agg_prediction)

            # Shift prediction timestamps if required
            if self.timestamp_shift != 0:
                shifted_agg_prediction = Annotation(agg_prediction.uri)
                for segment, track, speaker in agg_prediction.itertracks(
                    yield_label=True
                ):
                    new_segment = Segment(
                        segment.start + self.timestamp_shift,
                        segment.end + self.timestamp_shift,
                    )
                    shifted_agg_prediction[new_segment, track] = speaker
                agg_prediction = shifted_agg_prediction

            outputs.append((agg_prediction, agg_waveform))

            # Make place for new chunks in buffer if required
            if len(self.chunk_buffer) == self.pred_aggregation.num_overlapping_windows:
                self.chunk_buffer = self.chunk_buffer[1:]
                self.pred_buffer = self.pred_buffer[1:]

        return outputs



from abc import ABC, abstractmethod
from typing import Optional, Text

import rich
from rich.progress import Progress, TaskID
from tqdm import tqdm


class ProgressBar(ABC):
    @abstractmethod
    def create(
        self,
        total: int,
        description: Optional[Text] = None,
        unit: Text = "it",
        **kwargs,
    ):
        pass

    @abstractmethod
    def start(self):
        pass

    @abstractmethod
    def update(self, n: int = 1):
        pass

    @abstractmethod
    def write(self, text: Text):
        pass

    @abstractmethod
    def stop(self):
        pass

    @abstractmethod
    def close(self):
        pass

    @property
    @abstractmethod
    def default_description(self) -> Text:
        pass

    @property
    @abstractmethod
    def initial_description(self) -> Optional[Text]:
        pass

    def resolve_description(self, new_description: Optional[Text] = None) -> Text:
        if self.initial_description is None:
            if new_description is None:
                return self.default_description
            return new_description
        else:
            return self.initial_description


class RichProgressBar(ProgressBar):
    def __init__(
        self,
        description: Optional[Text] = None,
        color: Text = "green",
        leave: bool = True,
        do_close: bool = True,
    ):
        self.description = description
        self.color = color
        self.do_close = do_close
        self.bar = Progress(transient=not leave)
        self.bar.start()
        self.task_id: Optional[TaskID] = None

    @property
    def default_description(self) -> Text:
        return f"[{self.color}]Streaming"

    @property
    def initial_description(self) -> Optional[Text]:
        if self.description is not None:
            return f"[{self.color}]{self.description}"
        return self.description

    def create(
        self,
        total: int,
        description: Optional[Text] = None,
        unit: Text = "it",
        **kwargs,
    ):
        if self.task_id is None:
            self.task_id = self.bar.add_task(
                self.resolve_description(f"[{self.color}]{description}"),
                start=False,
                total=total,
                completed=0,
                visible=True,
                **kwargs,
            )

    def start(self):
        assert self.task_id is not None
        self.bar.start_task(self.task_id)

    def update(self, n: int = 1):
        assert self.task_id is not None
        self.bar.update(self.task_id, advance=n)

    def write(self, text: Text):
        rich.print(text)

    def stop(self):
        assert self.task_id is not None
        self.bar.stop_task(self.task_id)

    def close(self):
        if self.do_close:
            self.bar.stop()


class TQDMProgressBar(ProgressBar):
    def __init__(
        self,
        description: Optional[Text] = None,
        leave: bool = True,
        position: Optional[int] = None,
        do_close: bool = True,
    ):
        self.description = description
        self.leave = leave
        self.position = position
        self.do_close = do_close
        self.pbar: Optional[tqdm] = None

    @property
    def default_description(self) -> Text:
        return "Streaming"

    @property
    def initial_description(self) -> Optional[Text]:
        return self.description

    def create(
        self,
        total: int,
        description: Optional[Text] = None,
        unit: Optional[Text] = "it",
        **kwargs,
    ):
        if self.pbar is None:
            self.pbar = tqdm(
                desc=self.resolve_description(description),
                total=total,
                unit=unit,
                leave=self.leave,
                position=self.position,
                **kwargs,
            )

    def start(self):
        pass

    def update(self, n: int = 1):
        assert self.pbar is not None
        self.pbar.update(n)

    def write(self, text: Text):
        tqdm.write(text)

    def stop(self):
        self.close()

    def close(self):
        if self.do_close:
            assert self.pbar is not None
            self.pbar.close()



import argparse
from pathlib import Path

import pandas as pd
import torch

from diart import argdoc
from diart import models as m
from diart import utils
from diart.inference import Benchmark, Parallelize


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "root",
        type=Path,
        help="Directory with audio files CONVERSATION.(wav|flac|m4a|...)",
    )
    parser.add_argument(
        "--pipeline",
        default="SpeakerDiarization",
        type=str,
        help="Class of the pipeline to optimize. Defaults to 'SpeakerDiarization'",
    )
    parser.add_argument(
        "--segmentation",
        default="pyannote/segmentation",
        type=str,
        help=f"{argdoc.SEGMENTATION}. Defaults to pyannote/segmentation",
    )
    parser.add_argument(
        "--embedding",
        default="pyannote/embedding",
        type=str,
        help=f"{argdoc.EMBEDDING}. Defaults to pyannote/embedding",
    )
    parser.add_argument(
        "--reference",
        type=Path,
        help="Optional. Directory with RTTM files CONVERSATION.rttm. Names must match audio files",
    )
    parser.add_argument(
        "--duration",
        type=float,
        help=f"{argdoc.DURATION}. Defaults to training segmentation duration",
    )
    parser.add_argument(
        "--step", default=0.5, type=float, help=f"{argdoc.STEP}. Defaults to 0.5"
    )
    parser.add_argument(
        "--latency", default=0.5, type=float, help=f"{argdoc.LATENCY}. Defaults to 0.5"
    )
    parser.add_argument(
        "--tau-active", default=0.5, type=float, help=f"{argdoc.TAU}. Defaults to 0.5"
    )
    parser.add_argument(
        "--rho-update", default=0.3, type=float, help=f"{argdoc.RHO}. Defaults to 0.3"
    )
    parser.add_argument(
        "--delta-new", default=1, type=float, help=f"{argdoc.DELTA}. Defaults to 1"
    )
    parser.add_argument(
        "--gamma", default=3, type=float, help=f"{argdoc.GAMMA}. Defaults to 3"
    )
    parser.add_argument(
        "--beta", default=10, type=float, help=f"{argdoc.BETA}. Defaults to 10"
    )
    parser.add_argument(
        "--max-speakers",
        default=20,
        type=int,
        help=f"{argdoc.MAX_SPEAKERS}. Defaults to 20",
    )
    parser.add_argument(
        "--batch-size",
        default=32,
        type=int,
        help=f"{argdoc.BATCH_SIZE}. Defaults to 32",
    )
    parser.add_argument(
        "--num-workers",
        default=0,
        type=int,
        help=f"{argdoc.NUM_WORKERS}. Defaults to 0 (no parallelism)",
    )
    parser.add_argument(
        "--cpu",
        dest="cpu",
        action="store_true",
        help=f"{argdoc.CPU}. Defaults to GPU if available, CPU otherwise",
    )
    parser.add_argument(
        "--output", type=Path, help=f"{argdoc.OUTPUT}. Defaults to no writing"
    )
    parser.add_argument(
        "--hf-token",
        default="true",
        type=str,
        help=f"{argdoc.HF_TOKEN}. Defaults to 'true' (required by pyannote)",
    )
    args = parser.parse_args()

    # Resolve device
    args.device = torch.device("cpu") if args.cpu else None

    # Resolve models
    hf_token = utils.parse_hf_token_arg(args.hf_token)
    args.segmentation = m.SegmentationModel.from_pyannote(args.segmentation, hf_token)
    args.embedding = m.EmbeddingModel.from_pyannote(args.embedding, hf_token)

    pipeline_class = utils.get_pipeline_class(args.pipeline)

    benchmark = Benchmark(
        args.root,
        args.reference,
        args.output,
        show_progress=True,
        show_report=True,
        batch_size=args.batch_size,
    )

    config = pipeline_class.get_config_class()(**vars(args))
    if args.num_workers > 0:
        benchmark = Parallelize(benchmark, args.num_workers)

    report = benchmark(pipeline_class, config)

    if args.output is not None and isinstance(report, pd.DataFrame):
        report.to_csv(args.output / "benchmark_report.csv")


if __name__ == "__main__":
    run()



from typing import Optional, Union, Text

import torch
from einops import rearrange

from ..features import TemporalFeatures, TemporalFeatureFormatter
from ..models import EmbeddingModel


class SpeakerEmbedding:
    def __init__(self, model: EmbeddingModel, device: Optional[torch.device] = None):
        self.model = model
        self.model.eval()
        self.device = device
        if self.device is None:
            self.device = torch.device("cpu")
        self.model.to(self.device)
        self.waveform_formatter = TemporalFeatureFormatter()
        self.weights_formatter = TemporalFeatureFormatter()

    @staticmethod
    def from_pyannote(
        model,
        use_hf_token: Union[Text, bool, None] = True,
        device: Optional[torch.device] = None,
    ) -> "SpeakerEmbedding":
        emb_model = EmbeddingModel.from_pyannote(model, use_hf_token)
        return SpeakerEmbedding(emb_model, device)

    def __call__(
        self, waveform: TemporalFeatures, weights: Optional[TemporalFeatures] = None
    ) -> torch.Tensor:
        """
        Calculate speaker embeddings of input audio.
        If weights are given, calculate many speaker embeddings from the same waveform.

        Parameters
        ----------
        waveform: TemporalFeatures, shape (samples, channels) or (batch, samples, channels)
        weights: Optional[TemporalFeatures], shape (frames, speakers) or (batch, frames, speakers)
            Per-speaker and per-frame weights. Defaults to no weights.

        Returns
        -------
        embeddings: torch.Tensor
            If weights are provided, the shape is (batch, speakers, embedding_dim),
            otherwise the shape is (batch, embedding_dim).
            If batch size == 1, the batch dimension is omitted.
        """
        with torch.no_grad():
            inputs = self.waveform_formatter.cast(waveform).to(self.device)
            inputs = rearrange(inputs, "batch sample channel -> batch channel sample")
            if weights is not None:
                weights = self.weights_formatter.cast(weights).to(self.device)
                batch_size, _, num_speakers = weights.shape
                inputs = inputs.repeat(1, num_speakers, 1)
                weights = rearrange(weights, "batch frame spk -> (batch spk) frame")
                inputs = rearrange(inputs, "batch spk sample -> (batch spk) 1 sample")
                output = rearrange(
                    self.model(inputs, weights),
                    "(batch spk) feat -> batch spk feat",
                    batch=batch_size,
                    spk=num_speakers,
                )
            else:
                output = self.model(inputs)
            return output.squeeze().cpu()


class OverlappedSpeechPenalty:
    """
    Parameters
    ----------
    gamma: float, optional
        Exponent to lower low-confidence predictions.
        Defaults to 3.
    beta: float, optional
        Temperature parameter (actually 1/beta) to lower joint speaker activations.
        Defaults to 10.
    """

    def __init__(self, gamma: float = 3, beta: float = 10):
        self.gamma = gamma
        self.beta = beta
        self.formatter = TemporalFeatureFormatter()

    def __call__(self, segmentation: TemporalFeatures) -> TemporalFeatures:
        weights = self.formatter.cast(segmentation)  # shape (batch, frames, speakers)
        with torch.no_grad():
            probs = torch.softmax(self.beta * weights, dim=-1)
            weights = torch.pow(weights, self.gamma) * torch.pow(probs, self.gamma)
            weights[weights < 1e-8] = 1e-8
        return self.formatter.restore_type(weights)


class EmbeddingNormalization:
    def __init__(self, norm: Union[float, torch.Tensor] = 1):
        self.norm = norm
        # Add batch dimension if missing
        if isinstance(self.norm, torch.Tensor) and self.norm.ndim == 2:
            self.norm = self.norm.unsqueeze(0)

    def __call__(self, embeddings: torch.Tensor) -> torch.Tensor:
        # Add batch dimension if missing
        if embeddings.ndim == 2:
            embeddings = embeddings.unsqueeze(0)
        if isinstance(self.norm, torch.Tensor):
            batch_size1, num_speakers1, _ = self.norm.shape
            batch_size2, num_speakers2, _ = embeddings.shape
            assert batch_size1 == batch_size2 and num_speakers1 == num_speakers2
        with torch.no_grad():
            norm_embs = (
                self.norm
                * embeddings
                / torch.norm(embeddings, p=2, dim=-1, keepdim=True)
            )
        return norm_embs


class OverlapAwareSpeakerEmbedding:
    """
    Extract overlap-aware speaker embeddings given an audio chunk and its segmentation.

    Parameters
    ----------
    model: EmbeddingModel
        A pre-trained embedding model.
    gamma: float, optional
        Exponent to lower low-confidence predictions.
        Defaults to 3.
    beta: float, optional
        Softmax's temperature parameter (actually 1/beta) to lower joint speaker activations.
        Defaults to 10.
    norm: float or torch.Tensor of shape (batch, speakers, 1) where batch is optional
        The target norm for the embeddings. It can be different for each speaker.
        Defaults to 1.
    device: Optional[torch.device]
        The device on which to run the embedding model.
        Defaults to GPU if available or CPU if not.
    """

    def __init__(
        self,
        model: EmbeddingModel,
        gamma: float = 3,
        beta: float = 10,
        norm: Union[float, torch.Tensor] = 1,
        device: Optional[torch.device] = None,
    ):
        self.embedding = SpeakerEmbedding(model, device)
        self.osp = OverlappedSpeechPenalty(gamma, beta)
        self.normalize = EmbeddingNormalization(norm)

    @staticmethod
    def from_pyannote(
        model,
        gamma: float = 3,
        beta: float = 10,
        norm: Union[float, torch.Tensor] = 1,
        use_hf_token: Union[Text, bool, None] = True,
        device: Optional[torch.device] = None,
    ):
        model = EmbeddingModel.from_pyannote(model, use_hf_token)
        return OverlapAwareSpeakerEmbedding(model, gamma, beta, norm, device)

    def __call__(
        self, waveform: TemporalFeatures, segmentation: TemporalFeatures
    ) -> torch.Tensor:
        return self.normalize(self.embedding(waveform, self.osp(segmentation)))



SEGMENTATION = "Segmentation model name from pyannote"
EMBEDDING = "Embedding model name from pyannote"
DURATION = "Chunk duration (in seconds)"
STEP = "Sliding window step (in seconds)"
LATENCY = "System latency (in seconds). STEP <= LATENCY <= CHUNK_DURATION"
TAU = "Probability threshold to consider a speaker as active. 0 <= TAU <= 1"
RHO = "Speech ratio threshold to decide if centroids are updated with a given speaker. 0 <= RHO <= 1"
DELTA = "Embedding-to-centroid distance threshold to flag a speaker as known or new. 0 <= DELTA <= 2"
GAMMA = "Parameter gamma for overlapped speech penalty"
BETA = "Parameter beta for overlapped speech penalty"
MAX_SPEAKERS = "Maximum number of speakers"
CPU = "Force models to run on CPU"
BATCH_SIZE = "For segmentation and embedding pre-calculation. If BATCH_SIZE < 2, run fully online and estimate real-time latency"
NUM_WORKERS = "Number of parallel workers"
OUTPUT = "Directory to store the system's output in RTTM format"
HF_TOKEN = "Huggingface authentication token for hosted models ('true' | 'false' | <token>). If 'true', it will use the token from huggingface-cli login"
SAMPLE_RATE = "Sample rate of the audio stream"



from .aggregation import (
    AggregationStrategy,
    HammingWeightedAverageStrategy,
    AverageStrategy,
    FirstOnlyStrategy,
    DelayedAggregation,
)
from .clustering import OnlineSpeakerClustering
from .embedding import (
    SpeakerEmbedding,
    OverlappedSpeechPenalty,
    EmbeddingNormalization,
    OverlapAwareSpeakerEmbedding,
)
from .segmentation import SpeakerSegmentation
from .diarization import SpeakerDiarization, SpeakerDiarizationConfig
from .base import PipelineConfig, Pipeline
from .utils import Binarize, Resample, AdjustVolume
from .vad import VoiceActivityDetection, VoiceActivityDetectionConfig



from abc import ABC, abstractmethod
from typing import Optional, Text, Union, Callable

import torch
import torch.nn as nn

try:
    import pyannote.audio.pipelines.utils as pyannote_loader

    _has_pyannote = True
except ImportError:
    _has_pyannote = False


class PyannoteLoader:
    def __init__(self, model_info, hf_token: Union[Text, bool, None] = True):
        super().__init__()
        self.model_info = model_info
        self.hf_token = hf_token

    def __call__(self) -> nn.Module:
        return pyannote_loader.get_model(self.model_info, self.hf_token)


class LazyModel(nn.Module, ABC):
    def __init__(self, loader: Callable[[], nn.Module]):
        super().__init__()
        self.get_model = loader
        self.model: Optional[nn.Module] = None

    def is_in_memory(self) -> bool:
        """Return whether the model has been loaded into memory"""
        return self.model is not None

    def load(self):
        if not self.is_in_memory():
            self.model = self.get_model()

    def to(self, *args, **kwargs) -> nn.Module:
        self.load()
        return super().to(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        self.load()
        return super().__call__(*args, **kwargs)


class SegmentationModel(LazyModel):
    """
    Minimal interface for a segmentation model.
    """

    @staticmethod
    def from_pyannote(
        model, use_hf_token: Union[Text, bool, None] = True
    ) -> "SegmentationModel":
        """
        Returns a `SegmentationModel` wrapping a pyannote model.

        Parameters
        ----------
        model: pyannote.PipelineModel
            The pyannote.audio model to fetch.
        use_hf_token: str | bool, optional
            The Huggingface access token to use when downloading the model.
            If True, use huggingface-cli login token.
            Defaults to None.

        Returns
        -------
        wrapper: SegmentationModel
        """
        assert _has_pyannote, "No pyannote.audio installation found"
        return PyannoteSegmentationModel(model, use_hf_token)

    @property
    @abstractmethod
    def sample_rate(self) -> int:
        pass

    @property
    @abstractmethod
    def duration(self) -> float:
        pass

    @abstractmethod
    def forward(self, waveform: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the segmentation model.

        Parameters
        ----------
        waveform: torch.Tensor, shape (batch, channels, samples)

        Returns
        -------
        speaker_segmentation: torch.Tensor, shape (batch, frames, speakers)
        """
        pass


class PyannoteSegmentationModel(SegmentationModel):
    def __init__(self, model_info, hf_token: Union[Text, bool, None] = True):
        super().__init__(PyannoteLoader(model_info, hf_token))

    @property
    def sample_rate(self) -> int:
        self.load()
        return self.model.audio.sample_rate

    @property
    def duration(self) -> float:
        self.load()
        return self.model.specifications.duration

    def forward(self, waveform: torch.Tensor) -> torch.Tensor:
        return self.model(waveform)


class EmbeddingModel(LazyModel):
    """Minimal interface for an embedding model."""

    @staticmethod
    def from_pyannote(
        model, use_hf_token: Union[Text, bool, None] = True
    ) -> "EmbeddingModel":
        """
        Returns an `EmbeddingModel` wrapping a pyannote model.

        Parameters
        ----------
        model: pyannote.PipelineModel
            The pyannote.audio model to fetch.
        use_hf_token: str | bool, optional
            The Huggingface access token to use when downloading the model.
            If True, use huggingface-cli login token.
            Defaults to None.

        Returns
        -------
        wrapper: EmbeddingModel
        """
        assert _has_pyannote, "No pyannote.audio installation found"
        return PyannoteEmbeddingModel(model, use_hf_token)

    @abstractmethod
    def forward(
        self, waveform: torch.Tensor, weights: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        """
        Forward pass of an embedding model with optional weights.

        Parameters
        ----------
        waveform: torch.Tensor, shape (batch, channels, samples)
        weights: Optional[torch.Tensor], shape (batch, frames)
            Temporal weights for each sample in the batch. Defaults to no weights.

        Returns
        -------
        speaker_embeddings: torch.Tensor, shape (batch, embedding_dim)
        """
        pass


class PyannoteEmbeddingModel(EmbeddingModel):
    def __init__(self, model_info, hf_token: Union[Text, bool, None] = True):
        super().__init__(PyannoteLoader(model_info, hf_token))

    def forward(
        self,
        waveform: torch.Tensor,
        weights: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        return self.model(waveform, weights=weights)
